# SPDX-License-Identifier: GPL-3.0-or-later WITH GPL-3.0-linking-exception

import subprocess
import sys
import zipfile
from pathlib import Path

import pytest

import deluge.common
from deluge.pluginmanagerbase import _extract_egg, _PluginDirFinder


@pytest.fixture(scope='session')
def builtin_egg_info_dir(tmp_path_factory):
    """Session-scoped: generate egg-info for all built-in plugins into a tmp dir.

    Runs ``setup.py egg_info --egg-base <tmp/plugins>`` for each built-in plugin
    so the plugin scanner can discover them without polluting the source tree.

    Returns the plugins/ Path containing all generated .egg-info directories.
    """
    plugins_dir = tmp_path_factory.mktemp('builtin_plugins')

    plugin_src_root = Path(deluge.common.resource_filename('deluge', 'plugins'))
    for plugin_dir in sorted(plugin_src_root.iterdir()):
        if not (plugin_dir / 'setup.py').is_file():
            continue

        subprocess.run(
            [sys.executable, 'setup.py', 'egg_info', '--egg-base', str(plugins_dir)],
            cwd=str(plugin_dir),
            check=True,
            capture_output=True,
        )
    return plugins_dir


@pytest.fixture()
def fake_egg(tmp_path):
    """Create a minimal bdist_egg zip in a plugins directory.

    Includes a Python module (fake_plugin) with a resource file (config.ui)
    and entry points, so it can be used for both discovery and resource tests.

    Returns:
        plugins_dir, egg_path
    """
    pkg_info = 'Metadata-Version: 1.0\nName: FakePlugin\nVersion: 0.1\n'
    ep_txt = '[deluge.plugin.core]\nFakePlugin = fake_plugin:CorePlugin\n'
    plugins_dir = tmp_path / 'plugins'
    plugins_dir.mkdir()
    egg_path = plugins_dir / 'FakePlugin-0.1-py3.egg'

    with zipfile.ZipFile(egg_path, 'w') as zf:
        zf.writestr('EGG-INFO/PKG-INFO', pkg_info)
        zf.writestr('EGG-INFO/entry_points.txt', ep_txt)
        zf.writestr('fake_plugin/__init__.py', '')
        zf.writestr('fake_plugin/config.ui', '<ui/>')
        zf.writestr('fake_plugin/data/script.js', 'console.log("test");')

    return plugins_dir, egg_path


@pytest.fixture()
def installed_fake_plugin(fake_egg, tmp_path):
    """Extract fake_egg and register it via _PluginDirFinder on sys.meta_path.

    Yields with fake_plugin importable via importlib.resources, then cleans up.
    """
    _, egg_path = fake_egg
    dest = tmp_path / 'extracted'
    _extract_egg(egg_path, dest)

    finder = _PluginDirFinder(dest)
    sys.meta_path.insert(0, finder)
    # Evict any cached module so the import goes through our finder
    sys.modules.pop('fake_plugin', None)
    yield
    sys.meta_path.remove(finder)
    sys.modules.pop('fake_plugin', None)



#
# Copyright (C) 2010 Andrew Resch <andrewresch@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#
import logging
from time import time

from twisted.internet.defer import maybeDeferred, succeed

import deluge.component as component
from deluge.ui.client import client

log = logging.getLogger(__name__)


class SessionProxy(component.Component):
    """
    The SessionProxy component is used to cache session information client-side
    to reduce the number of RPCs needed to provide a rich user interface.

    It will query the Core for only changes in the status of the torrents
    and will try to satisfy client requests from the cache.

    """

    def __init__(self):
        log.debug('SessionProxy init..')
        component.Component.__init__(self, 'SessionProxy', interval=5)

        # Set the cache time in seconds
        # This is how long data will be valid before re-fetching from the core
        self.cache_time = 1.5

        # Hold the torrents' status.. {torrent_id: [time, {status_dict}], ...}
        self.torrents = {}

        # Holds the time of the last key update.. {torrent_id: {key1, time, ...}, ...}
        self.cache_times = {}

    def start(self):
        client.register_event_handler(
            'TorrentStateChangedEvent', self.on_torrent_state_changed
        )
        client.register_event_handler('TorrentRemovedEvent', self.on_torrent_removed)
        client.register_event_handler('TorrentAddedEvent', self.on_torrent_added)

        def on_get_session_state(torrent_ids):
            for torrent_id in torrent_ids:
                # Let's at least store the torrent ids with empty statuses
                # so that upcoming queries or status updates don't throw errors.
                self.torrents.setdefault(torrent_id, [time(), {}])
                self.cache_times.setdefault(torrent_id, {})
            return torrent_ids

        return client.core.get_session_state().addCallback(on_get_session_state)

    def stop(self):
        client.deregister_event_handler(
            'TorrentStateChangedEvent', self.on_torrent_state_changed
        )
        client.deregister_event_handler('TorrentRemovedEvent', self.on_torrent_removed)
        client.deregister_event_handler('TorrentAddedEvent', self.on_torrent_added)
        self.torrents = {}

    def create_status_dict(self, torrent_ids, keys):
        """
        Creates a status dict from the cache.

        :param torrent_ids: the torrent_ids
        :type torrent_ids: list of strings
        :param keys: the status keys
        :type keys: list of strings

        :returns: a dict with the status information for the *torrent_ids*
        :rtype: dict

        """
        sd = {}
        keys = set(keys)
        keys_len = (
            -1
        )  # The number of keys for the current cache (not the len of keys_diff_cached)
        keys_diff_cached = []

        for torrent_id in torrent_ids:
            try:
                if keys:
                    sd[torrent_id] = self.torrents[torrent_id][1].copy()

                    # Have to remove the keys that weren't requested
                    if len(sd[torrent_id]) == keys_len:
                        # If the number of keys are equal they are the same keys
                        # so we use the cached diff of the keys we need to remove
                        keys_to_remove = keys_diff_cached
                    else:
                        # Not the same keys so create a new diff
                        keys_to_remove = set(sd[torrent_id]) - keys
                        # Update the cached diff
                        keys_diff_cached = keys_to_remove
                        keys_len = len(sd[torrent_id])

                    # Usually there are no keys to remove, so it's cheaper with
                    # this if-test than a for-loop with no iterations.
                    if keys_to_remove:
                        for k in keys_to_remove:
                            del sd[torrent_id][k]
                else:
                    sd[torrent_id] = dict(self.torrents[torrent_id][1])
            except KeyError:
                continue
        return sd

    def get_torrent_status(self, torrent_id, keys):
        """
        Get a status dict for one torrent.

        :param torrent_id: the torrent_id
        :type torrent_id: string
        :param keys: the status keys
        :type keys: list of strings

        :returns: a dict of status information
        :rtype: dict

        """
        if torrent_id in self.torrents:
            # Keep track of keys we need to request from the core
            keys_to_get = []
            if not keys:
                keys = list(self.torrents[torrent_id][1])

            for key in keys:
                if (
                    time() - self.cache_times[torrent_id].get(key, 0.0)
                    > self.cache_time
                ):
                    keys_to_get.append(key)
            if not keys_to_get:
                return succeed(self.create_status_dict([torrent_id], keys)[torrent_id])
            else:
                d = client.core.get_torrent_status(torrent_id, keys_to_get, True)

                def on_status(result, torrent_id):
                    t = time()
                    try:
                        self.torrents[torrent_id][0] = t
                        self.torrents[torrent_id][1].update(result)
                        for key in keys_to_get:
                            self.cache_times[torrent_id][key] = t
                        return self.create_status_dict([torrent_id], keys)[torrent_id]
                    except KeyError:
                        log.debug(
                            f'Status missing for torrent (removed?): {torrent_id}'
                        )
                        return {}

                return d.addCallback(on_status, torrent_id)
        else:
            d = client.core.get_torrent_status(torrent_id, keys, True)

            def on_status(result):
                if result:
                    t = time()
                    self.torrents[torrent_id] = (t, result)
                    self.cache_times[torrent_id] = {}
                    for key in result:
                        self.cache_times[torrent_id][key] = t

                return result

            return d.addCallback(on_status)

    def get_torrents_status(self, filter_dict, keys):
        """
        Get a dict of torrent statuses.

        The filter can take 2 keys, *state* and *id*.  The state filter can be
        one of the torrent states or the special one *Active*.  The *id* key is
        simply a list of torrent_ids.

        :param filter_dict: the filter used for this query
        :type filter_dict: dict
        :param keys: the status keys
        :type keys: list of strings

        :returns: a dict of torrent_ids and their status dicts
        :rtype: dict

        """

        # Helper functions and callbacks ---------------------------------------
        def on_status(result, torrent_ids, keys):
            # Update the internal torrent status dict with the update values
            t = time()
            for key, value in result.items():
                try:
                    self.torrents[key][0] = t
                    self.torrents[key][1].update(value)
                    for k in value:
                        self.cache_times[key][k] = t
                except KeyError:
                    # The torrent was removed
                    continue

            # Create the status dict
            if not torrent_ids:
                torrent_ids = list(result)

            return self.create_status_dict(torrent_ids, keys)

        def find_torrents_to_fetch(torrent_ids):
            to_fetch = []
            t = time()
            for torrent_id in torrent_ids:
                torrent = self.torrents[torrent_id]
                if t - torrent[0] > self.cache_time:
                    to_fetch.append(torrent_id)
                else:
                    # We need to check if a key is expired
                    for key in keys:
                        if (
                            t - self.cache_times[torrent_id].get(key, 0.0)
                            > self.cache_time
                        ):
                            to_fetch.append(torrent_id)
                            break

            return to_fetch

        # -----------------------------------------------------------------------

        if not filter_dict:
            # This means we want all the torrents status
            # We get a list of any torrent_ids with expired status dicts
            torrents_list = list(self.torrents)
            to_fetch = find_torrents_to_fetch(torrents_list)
            if to_fetch:
                d = client.core.get_torrents_status({'id': to_fetch}, keys, True)
                return d.addCallback(on_status, torrents_list, keys)

            # Don't need to fetch anything
            return maybeDeferred(self.create_status_dict, torrents_list, keys)

        if len(filter_dict) == 1 and 'id' in filter_dict:
            # At this point we should have a filter with just "id" in it
            to_fetch = find_torrents_to_fetch(filter_dict['id'])
            if to_fetch:
                d = client.core.get_torrents_status({'id': to_fetch}, keys, True)
                return d.addCallback(on_status, filter_dict['id'], keys)
            else:
                # Don't need to fetch anything, so just return data from the cache
                return maybeDeferred(self.create_status_dict, filter_dict['id'], keys)
        else:
            # This is a keyworded filter so lets just pass it onto the core
            # XXX: Add more caching here.
            d = client.core.get_torrents_status(filter_dict, keys, True)
            return d.addCallback(on_status, None, keys)

    def on_torrent_state_changed(self, torrent_id, state):
        if torrent_id in self.torrents:
            self.torrents[torrent_id][1].setdefault('state', state)
            self.cache_times.setdefault(torrent_id, {}).update(state=time())

    def on_torrent_added(self, torrent_id, from_state):
        self.torrents[torrent_id] = [time() - self.cache_time - 1, {}]
        self.cache_times[torrent_id] = {}

        def on_status(status):
            self.torrents[torrent_id][1].update(status)
            t = time()
            for key in status:
                self.cache_times[torrent_id][key] = t

        client.core.get_torrent_status(torrent_id, []).addCallback(on_status)

    def on_torrent_removed(self, torrent_id):
        if torrent_id in self.torrents:
            del self.torrents[torrent_id]
            del self.cache_times[torrent_id]



#
# Copyright (C) 2016 bendikro <bro.devel+deluge@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import argparse
import shlex

import deluge.component as component
from deluge.ui.console.utils.colors import ConsoleColorFormatter


class OptionParserError(Exception):
    pass


class ConsoleBaseParser(argparse.ArgumentParser):
    def format_help(self):
        """Differs from ArgumentParser.format_help by adding the raw epilog
        as formatted in the string. Default behavior mangles the formatting.

        """
        # Handle epilog manually to keep the text formatting
        epilog = self.epilog
        self.epilog = ''
        help_str = super().format_help()
        if epilog is not None:
            help_str += epilog
        self.epilog = epilog
        return help_str


class ConsoleCommandParser(ConsoleBaseParser):
    def _split_args(self, args):
        command_options = []
        for a in args:
            if not a:
                continue
            if ';' in a:
                cmd_lines = [arg.strip() for arg in a.split(';')]
            elif ' ' in a:
                cmd_lines = [a]
            else:
                continue

            for cmd_line in cmd_lines:
                cmds = shlex.split(cmd_line)
                cmd_options = super().parse_args(args=cmds)
                cmd_options.command = cmds[0]
                command_options.append(cmd_options)

        return command_options

    def parse_args(self, args=None):
        """Parse known UI args and handle common and process group options.

        Notes:
            If started by deluge entry script this has already been done.

        Args:
            args (list, optional): The arguments to parse.

        Returns:
            argparse.Namespace: The parsed arguments.
        """
        from deluge.ui.ui_entry import AMBIGUOUS_CMD_ARGS

        self.base_parser.parse_known_ui_args(args, withhold=AMBIGUOUS_CMD_ARGS)

        multi_command = self._split_args(args)
        # If multiple commands were passed to console
        if multi_command:
            # With multiple commands, normal parsing will fail, so only parse
            # known arguments using the base parser, and then set
            # options.parsed_cmds to the already parsed commands
            options, remaining = self.base_parser.parse_known_args(args=args)
            options.parsed_cmds = multi_command
        else:
            subcommand = False
            if hasattr(self.base_parser, 'subcommand'):
                subcommand = getattr(self.base_parser, 'subcommand')
            if not subcommand:
                # We must use parse_known_args to handle case when no subcommand
                # is provided, because argparse does not support parsing without
                # a subcommand
                options, remaining = self.base_parser.parse_known_args(args=args)
                # If any options remain it means they do not exist. Reparse with
                # parse_args to trigger help message
                if remaining:
                    options = self.base_parser.parse_args(args=args)
                options.parsed_cmds = []
            else:
                options = super().parse_args(args=args)
                options.parsed_cmds = [options]

        if not hasattr(options, 'remaining'):
            options.remaining = []

        return options


class OptionParser(ConsoleBaseParser):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.formatter = ConsoleColorFormatter()

    def exit(self, status=0, msg=None):
        self._exit = True
        if msg:
            print(msg)

    def error(self, msg):
        """error(msg : string)

        Print a usage message incorporating 'msg' to stderr and exit.
        If you override this in a subclass, it should not return -- it
        should either exit or raise an exception.
        """
        raise OptionParserError(msg)

    def print_usage(self, _file=None):
        console = component.get('ConsoleUI')
        if self.usage:
            for line in self.format_usage().splitlines():
                console.write(line)

    def print_help(self, _file=None):
        console = component.get('ConsoleUI')
        console.set_batch_write(True)
        for line in self.format_help().splitlines():
            console.write(line)
        console.set_batch_write(False)

    def format_help(self):
        """Return help formatted with colors."""
        help_str = super().format_help()
        return self.formatter.format_colors(help_str)



#
# Copyright (C) 2016 bendikro <bro.devel+deluge@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

from deluge.decorators import overrides
from deluge.ui.console.utils import curses_util as util
from deluge.ui.console.utils.column import torrent_data_fields
from deluge.ui.console.widgets.fields import CheckedPlusInput, IntSpinInput
from deluge.ui.console.widgets.popup import InputPopup, MessagePopup

COLUMN_VIEW_HELP_STR = """
Control column visibilty with the following actions:

{!info!}'+'{!normal!} - {|indent_pos:|}Increase column width
{!info!}'-'{!normal!} - {|indent_pos:|}Decrease column width

{!info!}'CTRL+up'{!normal!} - {|indent_pos:|} Move column left
{!info!}'CTRL+down'{!normal!} - {|indent_pos:|} Move column right
"""

column_pref_names = [
    'queue',
    'name',
    'size',
    'downloaded',
    'uploaded',
    'remaining',
    'state',
    'progress',
    'seeds',
    'peers',
    'seeds_peers_ratio',
    'download_speed',
    'upload_speed',
    'max_download_speed',
    'max_upload_speed',
    'eta',
    'ratio',
    'avail',
    'time_added',
    'completed_time',
    'last_seen_complete',
    'tracker',
    'download_location',
    'active_time',
    'seeding_time',
    'finished_time',
    'time_since_transfer',
    'shared',
    'owner',
]


class ColumnAndWidth(CheckedPlusInput):
    def __init__(self, parent, name, message, child, on_width_func, **kwargs):
        CheckedPlusInput.__init__(self, parent, name, message, child, **kwargs)
        self.on_width_func = on_width_func

    @overrides(CheckedPlusInput)
    def handle_read(self, c):
        if c in [ord('+'), ord('-')]:
            val = self.child.get_value()
            change = 1 if chr(c) == '+' else -1
            self.child.set_value(val + change, validate=True)
            self.on_width_func(self.name, self.child.get_value())
            return util.ReadState.CHANGED
        return CheckedPlusInput.handle_read(self, c)


class TorrentViewColumns(InputPopup):
    def __init__(self, torrentlist):
        self.torrentlist = torrentlist
        self.torrentview = torrentlist.torrentview

        title = 'Visible columns (Esc to exit)'
        InputPopup.__init__(
            self,
            torrentlist,
            title,
            close_cb=self._do_set_column_visibility,
            immediate_action=True,
            height_req=len(column_pref_names) - 5,
            width_req=max(len(col) for col in column_pref_names + [title]) + 14,
            border_off_west=1,
            allow_rearrange=True,
        )

        msg_fmt = '%-25s'
        self.add_header((msg_fmt % _('Columns')) + ' ' + _('Width'), space_below=True)

        for colpref_name in column_pref_names:
            col = self.torrentview.config['torrentview']['columns'][colpref_name]
            width_spin = IntSpinInput(
                self,
                colpref_name + '_ width',
                '',
                self.move,
                col['width'],
                min_val=-1,
                max_val=99,
                fmt='%2d',
            )

            def on_width_func(name, width):
                self.torrentview.config['torrentview']['columns'][name]['width'] = width

            self._add_input(
                ColumnAndWidth(
                    self,
                    colpref_name,
                    torrent_data_fields[colpref_name]['name'],
                    width_spin,
                    on_width_func,
                    checked=col['visible'],
                    checked_char='*',
                    msg_fmt=msg_fmt,
                    show_usage_hints=False,
                    child_always_visible=True,
                )
            )

    def _do_set_column_visibility(
        self, data=None, state_changed=True, close=True, **kwargs
    ):
        if close:
            self.torrentlist.pop_popup()
            return
        elif not state_changed:
            return

        for key, value in data.items():
            self.torrentview.config['torrentview']['columns'][key]['visible'] = value[
                'value'
            ]
            self.torrentview.config['torrentview']['columns'][key]['order'] = value[
                'order'
            ]

        self.torrentview.config.save()
        self.torrentview.on_config_changed()
        self.torrentlist.refresh([])

    @overrides(InputPopup)
    def handle_read(self, c):
        if c == ord('h'):
            popup = MessagePopup(
                self.torrentlist,
                'Help',
                COLUMN_VIEW_HELP_STR,
                width_req=70,
                border_off_west=1,
            )
            self.torrentlist.push_popup(popup)
            return util.ReadState.READ
        return InputPopup.handle_read(self, c)



#
# Copyright (C) 2011 Nick Lanham <nick@afternight.org>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import logging

from twisted.internet import defer

import deluge.component as component
from deluge.common import TORRENT_STATE, fspeed
from deluge.ui.client import client

from . import BaseCommand

log = logging.getLogger(__name__)


class Command(BaseCommand):
    """Shows various status information from the daemon"""

    def add_arguments(self, parser):
        parser.add_argument(
            '-r',
            '--raw',
            action='store_true',
            default=False,
            dest='raw',
            help=_(
                'Raw values for upload/download rates (without KiB/s suffix)'
                '(useful for scripts that want to do their own parsing)'
            ),
        )
        parser.add_argument(
            '-n',
            '--no-torrents',
            action='store_false',
            default=True,
            dest='show_torrents',
            help=_('Do not show torrent status (Improves command speed)'),
        )

    def handle(self, options):
        self.console = component.get('ConsoleUI')
        self.status = None
        self.torrents = 1 if options.show_torrents else 0
        self.raw = options.raw

        def on_session_status(status):
            self.status = status

        def on_torrents_status(status):
            self.torrents = status

        def on_torrents_status_fail(reason):
            log.warning('Failed to retrieve session status: %s', reason)
            self.torrents = -2

        deferreds = []

        ds = client.core.get_session_status(
            [
                'peer.num_peers_connected',
                'payload_upload_rate',
                'payload_download_rate',
                'dht.dht_nodes',
            ]
        )
        ds.addCallback(on_session_status)
        deferreds.append(ds)

        if options.show_torrents:
            dt = client.core.get_torrents_status({}, ['state'])
            dt.addCallback(on_torrents_status)
            dt.addErrback(on_torrents_status_fail)
            deferreds.append(dt)

        return defer.DeferredList(deferreds).addCallback(self.print_status)

    def print_status(self, *args):
        self.console.set_batch_write(True)
        if self.raw:
            self.console.write(
                '{!info!}Total upload: %f' % self.status['payload_upload_rate']
            )
            self.console.write(
                '{!info!}Total download: %f' % self.status['payload_download_rate']
            )
        else:
            self.console.write(
                '{!info!}Total upload: %s' % fspeed(self.status['payload_upload_rate'])
            )
            self.console.write(
                '{!info!}Total download: %s'
                % fspeed(self.status['payload_download_rate'])
            )
        self.console.write('{!info!}DHT Nodes: %i' % self.status['dht.dht_nodes'])

        if isinstance(self.torrents, int):
            if self.torrents == -2:
                self.console.write('{!error!}Error getting torrent info')
        else:
            self.console.write('{!info!}Total torrents: %i' % len(self.torrents))
            state_counts = {}
            for state in TORRENT_STATE:
                state_counts[state] = 0
            for t in self.torrents:
                s = self.torrents[t]
                state_counts[s['state']] += 1
            for state in TORRENT_STATE:
                self.console.write('{!info!} %s: %i' % (state, state_counts[state]))

        self.console.set_batch_write(False)



#
# Copyright (C) 2011 Nick Lanham <nick@afternight.org>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import logging
from collections import deque

import deluge.component as component
from deluge.decorators import overrides
from deluge.ui.client import client
from deluge.ui.console.modes.basemode import BaseMode
from deluge.ui.console.modes.preferences.preference_panes import (
    BandwidthPane,
    CachePane,
    DaemonPane,
    DownloadsPane,
    InterfacePane,
    NetworkPane,
    OtherPane,
    ProxyPane,
    QueuePane,
)
from deluge.ui.console.utils import curses_util as util
from deluge.ui.console.widgets.fields import SelectInput
from deluge.ui.console.widgets.popup import MessagePopup, PopupsHandler
from deluge.ui.console.widgets.sidebar import Sidebar

try:
    import curses
except ImportError:
    pass


log = logging.getLogger(__name__)


# Big help string that gets displayed when the user hits 'h'
HELP_STR = """This screen lets you view and configure various options in deluge.

There are three main sections to this screen. Only one section is active at a time. \
You can switch the active section by hitting TAB (or Shift-TAB to go back one)

The section on the left displays the various categories that the settings fall in. \
You can navigate the list using the up/down arrows

The section on the right shows the settings for the selected category. When this \
section is active you can navigate the various settings with the up/down arrows. \
Special keys for each input type are described below.

The final section is at the bottom right, the: [Cancel] [Apply] [OK] buttons.
When this section is active, simply select the option you want using the arrow
keys and press Enter to confim.


Special keys for various input types are as follows:
- For text inputs you can simply type in the value.

{|indent:  |}- For numeric inputs (indicated by the value being in []s), you can type a value, \
or use PageUp and PageDown to increment/decrement the value.

- For checkbox inputs use the spacebar to toggle

{|indent:  |}- For checkbox plus something else inputs (the something else being only visible \
when you check the box) you can toggle the check with space, use the right \
arrow to edit the other value, and escape to get back to the check box.

"""


class ZONE:
    length = 3
    CATEGORIES, PREFRENCES, ACTIONS = list(range(length))


class PreferenceSidebar(Sidebar):
    def __init__(self, torrentview, width):
        height = curses.LINES - 2
        Sidebar.__init__(
            self, torrentview, width, height, title=None, border_off_north=1
        )
        self.categories = [
            _('Interface'),
            _('Downloads'),
            _('Network'),
            _('Bandwidth'),
            _('Other'),
            _('Daemon'),
            _('Queue'),
            _('Proxy'),
            _('Cache'),
        ]
        for name in self.categories:
            self.add_text_field(
                name,
                name,
                selectable=True,
                font_unfocused_active='bold',
                color_unfocused_active='white,black',
            )

    def on_resize(self):
        self.resize_window(curses.LINES - 2, self.width)


class Preferences(BaseMode, PopupsHandler):
    def __init__(self, parent_mode, stdscr, console_config, encoding=None):
        BaseMode.__init__(self, stdscr, encoding=encoding, do_refresh=False)
        PopupsHandler.__init__(self)
        self.parent_mode = parent_mode
        self.cur_cat = 0
        self.messages = deque()
        self.action_input = None
        self.config_loaded = False
        self.console_config = console_config
        self.active_port = -1
        self.active_zone = ZONE.CATEGORIES
        self.sidebar_width = 15  # Width of the categories pane

        self.sidebar = PreferenceSidebar(parent_mode, self.sidebar_width)
        self.sidebar.set_focused(True)
        self.sidebar.active_input = 0

        self._calc_sizes(resize=False)

        self.panes = [
            InterfacePane(self),
            DownloadsPane(self),
            NetworkPane(self),
            BandwidthPane(self),
            OtherPane(self),
            DaemonPane(self),
            QueuePane(self),
            ProxyPane(self),
            CachePane(self),
        ]

        self.action_input = SelectInput(
            self,
            None,
            None,
            [_('Cancel'), _('Apply'), _('OK')],
            [0, 1, 2],
            0,
            require_select_action=False,
        )

    def load_config(self):
        if self.config_loaded:
            return

        def on_get_config(core_config):
            self.core_config = core_config
            self.config_loaded = True
            for p in self.panes:
                p.create_pane(core_config, self.console_config)
            self.refresh()

        client.core.get_config().addCallback(on_get_config)

        def on_get_listen_port(port):
            self.active_port = port

        client.core.get_listen_port().addCallback(on_get_listen_port)

    @property
    def height(self):
        # top/bottom bars: 2, Action buttons (Cancel/Apply/OK): 1
        return self.rows - 3

    @property
    def width(self):
        return self.prefs_width

    def _calc_sizes(self, resize=True):
        self.prefs_width = self.cols - self.sidebar_width

        if not resize:
            return

        for p in self.panes:
            p.resize_window(self.height, p.pane_width)

    def _draw_preferences(self):
        self.cur_cat = self.sidebar.active_input
        self.panes[self.cur_cat].render(
            self, self.stdscr, self.prefs_width, self.active_zone == ZONE.PREFRENCES
        )
        self.panes[self.cur_cat].refresh()

    def _draw_actions(self):
        selected = self.active_zone == ZONE.ACTIONS
        self.stdscr.hline(self.rows - 3, self.sidebar_width, b'_', self.cols)
        self.action_input.render(
            self.stdscr,
            self.rows - 2,
            width=self.cols,
            active=selected,
            focus=True,
            col=self.cols - 22,
        )

    @overrides(BaseMode)
    def on_resize(self, rows, cols):
        BaseMode.on_resize(self, rows, cols)
        self._calc_sizes()

        if self.popup:
            self.popup.handle_resize()

        self.sidebar.on_resize()
        self.refresh()

    @overrides(component.Component)
    def update(self):
        for i, p in enumerate(self.panes):
            self.panes[i].update(i == self.cur_cat)

    @overrides(BaseMode)
    def resume(self):
        BaseMode.resume(self)
        self.sidebar.show()

    @overrides(BaseMode)
    def refresh(self):
        if (
            not component.get('ConsoleUI').is_active_mode(self)
            or not self.config_loaded
        ):
            return

        if self.popup is None and self.messages:
            title, msg = self.messages.popleft()
            self.push_popup(MessagePopup(self, title, msg))

        self.stdscr.erase()
        self.draw_statusbars()
        self._draw_actions()
        # Necessary to force updating the stdscr
        self.stdscr.noutrefresh()

        self.sidebar.refresh()

        # do this last since it moves the cursor
        self._draw_preferences()

        if self.popup:
            self.popup.refresh()

        curses.doupdate()

    def _apply_prefs(self):
        if self.core_config is None:
            return

        def update_conf_value(key, source_dict, dest_dict, updated):
            if dest_dict[key] != source_dict[key]:
                dest_dict[key] = source_dict[key]
                updated = True
            return updated

        new_core_config = {}
        for pane in self.panes:
            if not isinstance(pane, InterfacePane):
                pane.add_config_values(new_core_config)
        # Apply Core Prefs
        if client.connected():
            # Only do this if we're connected to a daemon
            config_to_set = {}
            for key in new_core_config:
                # The values do not match so this needs to be updated
                if self.core_config[key] != new_core_config[key]:
                    config_to_set[key] = new_core_config[key]

            if config_to_set:
                # Set each changed config value in the core
                client.core.set_config(config_to_set)
                client.force_call(True)
                # Update the configuration
                self.core_config.update(config_to_set)

        # Update Interface Prefs
        new_console_config = {}
        didupdate = False
        for pane in self.panes:
            # could just access panes by index, but that would break if panes
            # are ever reordered, so do it the slightly slower but safer way
            if isinstance(pane, InterfacePane):
                pane.add_config_values(new_console_config)
                for k in ['ring_bell', 'language']:
                    didupdate = update_conf_value(
                        k, new_console_config, self.console_config, didupdate
                    )
                for k in ['separate_complete', 'move_selection']:
                    didupdate = update_conf_value(
                        k,
                        new_console_config,
                        self.console_config['torrentview'],
                        didupdate,
                    )
                for k in [
                    'ignore_duplicate_lines',
                    'save_command_history',
                    'third_tab_lists_all',
                    'torrents_per_tab_press',
                ]:
                    didupdate = update_conf_value(
                        k, new_console_config, self.console_config['cmdline'], didupdate
                    )

        if didupdate:
            self.parent_mode.on_config_changed()

    def _update_preferences(self, core_config, console_config=None):
        self.core_config = core_config
        for pane in self.panes:
            if isinstance(pane, InterfacePane) and console_config:
                pane.update_values(console_config)
            else:
                pane.update_values(core_config)

    def _actions_read(self, c):
        self.action_input.handle_read(c)
        if c in [curses.KEY_ENTER, util.KEY_ENTER2]:
            # take action
            if self.action_input.selected_index == 0:  # Cancel
                # Reload stored config for panes
                self._update_preferences(self.core_config, self.console_config)
                self.back_to_parent()
            elif self.action_input.selected_index == 1:  # Apply
                self._apply_prefs()
                client.core.get_config().addCallback(self._update_preferences)
            elif self.action_input.selected_index == 2:  # OK
                self._apply_prefs()
                self.back_to_parent()

    def back_to_parent(self):
        component.get('ConsoleUI').set_mode(self.parent_mode.mode_name)

    @overrides(BaseMode)
    def read_input(self):
        c = self.stdscr.getch()

        if self.popup:
            if self.popup.handle_read(c):
                self.pop_popup()
            self.refresh()
            return

        if util.is_printable_chr(c):
            char = chr(c)
            if char == 'Q':
                component.get('ConsoleUI').quit()
            elif char == 'h':
                self.push_popup(MessagePopup(self, 'Preferences Help', HELP_STR))

        if self.sidebar.has_focus() and c == util.KEY_ESC:
            self.back_to_parent()
            return

        def update_active_zone(val):
            self.active_zone += val
            if self.active_zone == -1:
                self.active_zone = ZONE.length - 1
            else:
                self.active_zone %= ZONE.length
            self.sidebar.set_focused(self.active_zone == ZONE.CATEGORIES)

        if c == util.KEY_TAB:
            update_active_zone(1)
        elif c == curses.KEY_BTAB:
            update_active_zone(-1)
        else:
            if self.active_zone == ZONE.CATEGORIES:
                self.sidebar.handle_read(c)
            elif self.active_zone == ZONE.PREFRENCES:
                self.panes[self.cur_cat].handle_read(c)
            elif self.active_zone == ZONE.ACTIONS:
                self._actions_read(c)

        self.refresh()

    def is_active_pane(self, pane):
        return pane == self.panes[self.cur_cat]



def migrate_1_to_2(config):
    """Create better structure by moving most settings out of dict root
    and into sub categories. Some keys are also renamed to be consistent
    with other UIs.
    """

    def move_key(source, dest, source_key, dest_key=None):
        if dest_key is None:
            dest_key = source_key

        dest[dest_key] = source[source_key]
        del source[source_key]

    # These are moved to 'torrentview' sub dict
    for k in [
        'sort_primary',
        'sort_secondary',
        'move_selection',
        'separate_complete',
    ]:
        move_key(config, config['torrentview'], k)

    # These are moved to 'addtorrents' sub dict
    for k in [
        'show_misc_files',
        'show_hidden_folders',
        'sort_column',
        'reverse_sort',
        'last_path',
    ]:
        move_key(config, config['addtorrents'], 'addtorrents_%s' % k, dest_key=k)

    # These are moved to 'cmdline' sub dict
    for k in [
        'ignore_duplicate_lines',
        'torrents_per_tab_press',
        'third_tab_lists_all',
    ]:
        move_key(config, config['cmdline'], k)

    move_key(
        config,
        config['cmdline'],
        'save_legacy_history',
        dest_key='save_command_history',
    )

    # Add key for localization
    config['language'] = ''

    # Migrate column settings
    columns = [
        'queue',
        'size',
        'state',
        'progress',
        'seeds',
        'peers',
        'downspeed',
        'upspeed',
        'eta',
        'ratio',
        'avail',
        'added',
        'tracker',
        'savepath',
        'downloaded',
        'uploaded',
        'remaining',
        'owner',
        'downloading_time',
        'seeding_time',
        'completed',
        'seeds_peers_ratio',
        'complete_seen',
        'down_limit',
        'up_limit',
        'shared',
        'name',
    ]
    column_name_mapping = {
        'downspeed': 'download_speed',
        'upspeed': 'upload_speed',
        'added': 'time_added',
        'savepath': 'download_location',
        'completed': 'completed_time',
        'complete_seen': 'last_seen_complete',
        'down_limit': 'max_download_speed',
        'up_limit': 'max_upload_speed',
        'downloading_time': 'active_time',
    }

    from deluge.ui.console.modes.torrentlist.torrentview import default_columns

    # These are moved to 'torrentview.columns' sub dict
    for k in columns:
        column_name = column_name_mapping.get(k, k)
        config['torrentview']['columns'][column_name] = {}
        if k == 'name':
            config['torrentview']['columns'][column_name]['visible'] = True
        else:
            move_key(
                config,
                config['torrentview']['columns'][column_name],
                'show_%s' % k,
                dest_key='visible',
            )
        move_key(
            config,
            config['torrentview']['columns'][column_name],
            '%s_width' % k,
            dest_key='width',
        )
        config['torrentview']['columns'][column_name]['order'] = default_columns[
            column_name
        ]['order']

    return config



#
# Copyright (C) 2009-2010 Damien Churchill <damoxc@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import email.message
import json
import logging
import os
import shutil
import tempfile
from base64 import b64encode
from types import FunctionType

from twisted.internet import defer, reactor
from twisted.internet.defer import Deferred, DeferredList
from twisted.web import http, resource, server

from deluge import component, httpdownloader
from deluge.common import AUTH_LEVEL_DEFAULT, get_magnet_info, is_magnet
from deluge.configmanager import get_config_dir
from deluge.decorators import maybe_coroutine
from deluge.error import NotAuthorizedError
from deluge.i18n import get_languages
from deluge.ui.client import client
from deluge.ui.common import FileTree2, TorrentInfo
from deluge.ui.coreconfig import CoreConfig
from deluge.ui.hostlist import HostList
from deluge.ui.sessionproxy import SessionProxy

log = logging.getLogger(__name__)


class JSONComponent(component.Component):
    def __init__(self, name, interval=1, depend=None):
        super().__init__(name, interval, depend)
        self._json = component.get('JSON')
        self._json.register_object(self, name)


def export(auth_level=AUTH_LEVEL_DEFAULT):
    """
    Decorator function to register an object's method as a RPC. The object
    will need to be registered with a `:class:JSON` to be effective.

    :param func: the function to export
    :type func: function
    :param auth_level: the auth level required to call this method
    :type auth_level: int

    """

    def wrap(func, *args, **kwargs):
        func._json_export = True
        func._json_auth_level = auth_level
        return func

    if isinstance(auth_level, FunctionType):
        func = auth_level
        auth_level = AUTH_LEVEL_DEFAULT
        return wrap(func)
    else:
        return wrap


class JSONException(Exception):
    def __init__(self, inner_exception):
        self.inner_exception = inner_exception
        Exception.__init__(self, str(inner_exception))


class JSON(resource.Resource, component.Component):
    """
    A Twisted Web resource that exposes a JSON-RPC interface for web clients \
    to use.
    """

    def __init__(self):
        resource.Resource.__init__(self)
        component.Component.__init__(self, 'JSON')
        self._remote_methods = []
        self._local_methods = {}
        if client.is_standalone():
            self.get_remote_methods()

    def get_remote_methods(self, result=None):
        """
        Updates remote methods from the daemon.

        Returns:
            t.i.d.Deferred: A deferred returning the available remote methods
        """

        def on_get_methods(methods):
            self._remote_methods = methods
            return methods

        return client.daemon.get_method_list().addCallback(on_get_methods)

    def _exec_local(self, method, params, request):
        """
        Handles executing all local methods.
        """
        if method == 'system.listMethods':
            d = Deferred()
            methods = list(self._remote_methods)
            methods.extend(self._local_methods)
            d.callback(methods)
            return d
        elif method in self._local_methods:
            # This will eventually process methods that the server adds
            # and any plugins.
            meth = self._local_methods[method]
            meth.__globals__['__request__'] = request
            component.get('Auth').check_request(request, meth)
            return meth(*params)
        raise JSONException('Unknown system method')

    def _exec_remote(self, method, params, request):
        """
        Executes methods using the Deluge client.
        """
        component.get('Auth').check_request(request, level=AUTH_LEVEL_DEFAULT)
        core_component, method = method.split('.')
        return getattr(getattr(client, core_component), method)(*params)

    def _handle_request(self, request):
        """
        Takes some json data as a string and attempts to decode it, and process
        the rpc object that should be contained, returning a deferred for all
        procedure calls and the request id.
        """
        try:
            request_data = json.loads(request.json.decode())
        except (ValueError, TypeError):
            raise JSONException('JSON not decodable')

        try:
            method = request_data['method']
            params = request_data['params']
            request_id = request_data['id']
        except KeyError as ex:
            message = 'Invalid JSON request, missing param {} in {}'.format(
                ex,
                request_data,
            )
            raise JSONException(message)

        result = None
        error = None

        try:
            if method.startswith('system.') or method in self._local_methods:
                result = self._exec_local(method, params, request)
            elif method in self._remote_methods:
                result = self._exec_remote(method, params, request)
            else:
                error = {'message': 'Unknown method', 'code': 2}
        except NotAuthorizedError:
            error = {'message': 'Not authenticated', 'code': 1}
        except Exception as ex:
            log.error('Error calling method `%s`: %s', method, ex)
            log.exception(ex)
            error = {'message': f'{ex.__class__.__name__}: {str(ex)}', 'code': 3}

        return request_id, result, error

    def _on_rpc_request_finished(self, result, response, request):
        """
        Sends the response of any rpc calls back to the json-rpc client.
        """
        response['result'] = result
        return self._send_response(request, response)

    def _on_rpc_request_failed(self, reason, response, request):
        """
        Handles any failures that occurred while making an rpc call.
        """
        log.error(reason)
        response['error'] = {
            'message': f'{reason.__class__.__name__}: {str(reason)}',
            'code': 4,
        }
        return self._send_response(request, response)

    def _on_json_request(self, request):
        """
        Handler to take the json data as a string and pass it on to the
        _handle_request method for further processing.
        """
        message = email.message.EmailMessage()
        message['content-type'] = request.getHeader(b'content-type').decode()
        content_type = message.get_content_type()
        if content_type != 'application/json':
            message = 'Invalid JSON request content-type: %s' % content_type
            raise JSONException(message)

        log.debug('json-request: %s', request.json)
        response = {'result': None, 'error': None, 'id': None}
        response['id'], d, response['error'] = self._handle_request(request)

        if isinstance(d, Deferred):
            d.addCallback(self._on_rpc_request_finished, response, request)
            d.addErrback(self._on_rpc_request_failed, response, request)
            return d
        else:
            response['result'] = d
            return self._send_response(request, response)

    def _on_json_request_failed(self, reason, request):
        """
        Returns the error in json response.
        """
        log.error(reason)
        response = {
            'result': None,
            'id': None,
            'error': {
                'code': 5,
                'message': f'{reason.__class__.__name__}: {str(reason)}',
            },
        }
        return self._send_response(request, response)

    def _send_response(self, request, response):
        if request._disconnected:
            return ''
        response = json.dumps(response)
        request.setHeader(b'content-type', b'application/json')
        request.write(response.encode())
        request.finish()
        return server.NOT_DONE_YET

    def render(self, request):
        """
        Handles all the POST requests made to the /json controller.
        """
        if request.method != b'POST':
            request.setResponseCode(http.NOT_ALLOWED)
            request.finish()
            return server.NOT_DONE_YET

        try:
            request.content.seek(0)
            request.json = request.content.read()
            self._on_json_request(request)
            return server.NOT_DONE_YET
        except Exception as ex:
            return self._on_json_request_failed(ex, request)

    def register_object(self, obj, name=None):
        """Registers an object to export it's rpc methods.

        These methods should be exported with the export decorator prior
        to registering the object.

        Args:
            obj (object): The object that we want to export.
            name (str): The name to use. If None, uses the object class name.

        """
        name = name or obj.__class__.__name__
        name = name.lower()

        for d in dir(obj):
            if d[0] == '_':
                continue
            if getattr(getattr(obj, d), '_json_export', False):
                log.debug('Registering method: %s', name + '.' + d)
                self._local_methods[name + '.' + d] = getattr(obj, d)

    def deregister_object(self, obj):
        """Deregisters an objects exported rpc methods.

        Args:
            obj (object): The object that was previously registered.

        """
        for key, value in self._local_methods.items():
            if value.__self__ == obj:
                del self._local_methods[key]


FILES_KEYS = ['files', 'file_progress', 'file_priorities']


class EventQueue:
    """
    This class subscribes to events from the core and stores them until all
    the subscribed listeners have received the events.
    """

    def __init__(self):
        self.__events = {}
        self.__handlers = {}
        self.__queue = {}
        self.__requests = {}

    def add_listener(self, listener_id, event):
        """
        Add a listener to the event queue.

        :param listener_id: A unique id for the listener
        :type listener_id: string
        :param event: The event name
        :type event: string
        """
        if event not in self.__events:

            def on_event(*args):
                for listener in self.__events[event]:
                    if listener not in self.__queue:
                        self.__queue[listener] = []
                    self.__queue[listener].append((event, args))

            client.register_event_handler(event, on_event)
            self.__handlers[event] = on_event
            self.__events[event] = [listener_id]
        elif listener_id not in self.__events[event]:
            self.__events[event].append(listener_id)

    def get_events(self, listener_id):
        """
        Retrieve the pending events for the listener.

        :param listener_id: A unique id for the listener
        :type listener_id: string
        """

        # Check to see if we have anything to return immediately
        if listener_id in self.__queue:
            queue = self.__queue[listener_id]
            del self.__queue[listener_id]
            return queue

        # Create a deferred to and check again in 100ms
        d = Deferred()
        reactor.callLater(0.1, self._get_events, listener_id, 0, d)
        return d

    def _get_events(self, listener_id, count, d):
        if listener_id in self.__queue:
            queue = self.__queue[listener_id]
            del self.__queue[listener_id]
            d.callback(queue)
        else:
            # Prevent this loop going on indefinitely incase a client leaves
            # the page or disconnects uncleanly.
            if count >= 50:
                d.callback(None)
            else:
                reactor.callLater(0.1, self._get_events, listener_id, count + 1, d)

    def remove_listener(self, listener_id, event):
        """
        Remove a listener from the event queue.

        :param listener_id: The unique id for the listener
        :type listener_id: string
        :param event: The event name
        :type event: string
        """
        self.__events[event].remove(listener_id)
        if not self.__events[event]:
            client.deregister_event_handler(event, self.__handlers[event])
            del self.__events[event]
            del self.__handlers[event]


class WebApi(JSONComponent):
    """
    The component that implements all the methods required for managing
    the web interface. The complete web json interface also exposes all the
    methods available from the core RPC.
    """

    def __init__(self):
        super().__init__('Web', depend=['SessionProxy'])
        self.hostlist = HostList()
        self.core_config = CoreConfig()
        self.event_queue = EventQueue()
        try:
            self.sessionproxy = component.get('SessionProxy')
        except KeyError:
            self.sessionproxy = SessionProxy()

    def disable(self):
        client.deregister_event_handler(
            'PluginEnabledEvent', self._json.get_remote_methods
        )
        client.deregister_event_handler(
            'PluginDisabledEvent', self._json.get_remote_methods
        )

        if client.is_standalone():
            component.get('Web.PluginManager').stop()
        else:
            client.disconnect()
            client.set_disconnect_callback(None)

    def enable(self):
        client.register_event_handler(
            'PluginEnabledEvent', self._json.get_remote_methods
        )
        client.register_event_handler(
            'PluginDisabledEvent', self._json.get_remote_methods
        )

        if client.is_standalone():
            component.get('Web.PluginManager').start()
        else:
            client.set_disconnect_callback(self._on_client_disconnect)
            default_host_id = component.get('DelugeWeb').config['default_daemon']
            if default_host_id:
                return self.connect(default_host_id)

        return defer.succeed(True)

    def _on_client_connect(self, *args):
        """Handles client successfully connecting to the daemon.

        Invokes retrieving the method names and starts webapi and plugins.

        """
        d_methods = self._json.get_remote_methods()
        component.get('Web.PluginManager').start()
        self.start()
        return d_methods

    def _on_client_connect_fail(self, result, host_id):
        log.error(
            'Unable to connect to daemon, check host_id "%s" is correct.', host_id
        )

    def _on_client_disconnect(self, *args):
        component.get('Web.PluginManager').stop()
        return self.stop()

    def start(self):
        self.core_config.start()
        return self.sessionproxy.start()

    def stop(self):
        self.core_config.stop()
        self.sessionproxy.stop()
        return defer.succeed(True)

    @export
    def connect(self, host_id):
        """Connect the web client to a daemon.

        Args:
            host_id (str): The id of the daemon in the host list.

        Returns:
            Deferred: List of methods the daemon supports.
        """
        d = self.hostlist.connect_host(host_id)
        d.addCallback(self._on_client_connect)
        d.addErrback(self._on_client_connect_fail, host_id)
        return d

    @export
    def connected(self):
        """
        The current connection state.

        :returns: True if the client is connected
        :rtype: boolean
        """
        return client.connected()

    @export
    def disconnect(self):
        """
        Disconnect the web interface from the connected daemon.
        """
        d = client.disconnect()

        def on_disconnect(reason):
            return str(reason)

        d.addCallback(on_disconnect)
        return d

    @export
    def update_ui(self, keys, filter_dict):
        """
        Gather the information required for updating the web interface.

        :param keys: the information about the torrents to gather
        :type keys: list
        :param filter_dict: the filters to apply when selecting torrents.
        :type filter_dict: dictionary
        :returns: The torrent and UI information.
        :rtype: dictionary
        """
        d = Deferred()
        ui_info = {
            'connected': client.connected(),
            'torrents': None,
            'filters': None,
            'stats': {
                'max_download': self.core_config.get('max_download_speed'),
                'max_upload': self.core_config.get('max_upload_speed'),
                'max_num_connections': self.core_config.get('max_connections_global'),
            },
        }

        if not client.connected():
            d.callback(ui_info)
            return d

        def got_stats(stats):
            ui_info['stats']['num_connections'] = stats['peer.num_peers_connected']
            ui_info['stats']['upload_rate'] = stats['payload_upload_rate']
            ui_info['stats']['download_rate'] = stats['payload_download_rate']
            ui_info['stats']['download_protocol_rate'] = (
                stats['download_rate'] - stats['payload_download_rate']
            )
            ui_info['stats']['upload_protocol_rate'] = (
                stats['upload_rate'] - stats['payload_upload_rate']
            )
            ui_info['stats']['dht_nodes'] = stats['dht.dht_nodes']
            ui_info['stats']['has_incoming_connections'] = stats[
                'net.has_incoming_connections'
            ]

        def got_filters(filters):
            ui_info['filters'] = filters

        def got_free_space(free_space):
            ui_info['stats']['free_space'] = free_space

        def got_external_ip(external_ip):
            ui_info['stats']['external_ip'] = external_ip

        def got_torrents(torrents):
            ui_info['torrents'] = torrents

        def on_complete(result):
            d.callback(ui_info)

        d1 = component.get('SessionProxy').get_torrents_status(filter_dict, keys)
        d1.addCallback(got_torrents)

        d2 = client.core.get_filter_tree()
        d2.addCallback(got_filters)

        d3 = client.core.get_session_status(
            [
                'peer.num_peers_connected',
                'payload_download_rate',
                'payload_upload_rate',
                'download_rate',
                'upload_rate',
                'dht.dht_nodes',
                'net.has_incoming_connections',
            ]
        )
        d3.addCallback(got_stats)

        d4 = client.core.get_free_space(self.core_config.get('download_location'))
        d4.addCallback(got_free_space)

        d5 = client.core.get_external_ip()
        d5.addCallback(got_external_ip)

        dl = DeferredList([d1, d2, d3, d4, d5], consumeErrors=True)
        dl.addCallback(on_complete)
        return d

    def _on_got_files(self, torrent, d):
        files = torrent.get('files')
        file_progress = torrent.get('file_progress')
        file_priorities = torrent.get('file_priorities')

        paths = []
        info = {}
        for index, torrent_file in enumerate(files):
            path = torrent_file['path']
            paths.append(path)
            torrent_file['progress'] = file_progress[index]
            torrent_file['priority'] = file_priorities[index]
            torrent_file['index'] = index
            torrent_file['path'] = path
            info[path] = torrent_file

            # update the directory info
            dirname = os.path.dirname(path)
            while dirname:
                dirinfo = info.setdefault(dirname, {})
                dirinfo['size'] = dirinfo.get('size', 0) + torrent_file['size']
                if 'priority' not in dirinfo:
                    dirinfo['priority'] = torrent_file['priority']
                else:
                    if dirinfo['priority'] != torrent_file['priority']:
                        dirinfo['priority'] = 9

                progresses = dirinfo.setdefault('progresses', [])
                progresses.append(torrent_file['size'] * torrent_file['progress'] / 100)
                if dirinfo['size'] > 0:
                    dirinfo['progress'] = sum(progresses) / dirinfo['size'] * 100
                else:
                    dirinfo['progress'] = 100
                dirinfo['path'] = dirname
                dirname = os.path.dirname(dirname)

        def walk(path, item):
            if item['type'] == 'dir':
                item.update(info[path])
                return item
            else:
                item.update(info[path])
                return item

        file_tree = FileTree2(paths)
        file_tree.walk(walk)
        d.callback(file_tree.get_tree())

    @export
    def get_torrent_status(self, torrent_id, keys):
        """Get the status for a torrent, filtered by status keys."""
        return component.get('SessionProxy').get_torrent_status(torrent_id, keys)

    @export
    def get_torrent_files(self, torrent_id):
        """
        Gets the files for a torrent in tree format

        :param torrent_id: the id of the torrent to retrieve.
        :type torrent_id: string
        :returns: The torrents files in a tree
        :rtype: dictionary
        """
        main_deferred = Deferred()
        d = component.get('SessionProxy').get_torrent_status(torrent_id, FILES_KEYS)
        d.addCallback(self._on_got_files, main_deferred)
        return main_deferred

    @export
    def download_torrent_from_url(self, url, cookie=None):
        """
        Download a torrent file from a URL to a temporary directory.

        :param url: the URL of the torrent
        :type url: string
        :returns: the temporary file name of the torrent file
        :rtype: string
        """

        def on_download_success(result):
            log.debug('Successfully downloaded %s to %s', url, result)
            return result

        def on_download_fail(result):
            log.error('Failed to add torrent from url %s', url)
            return result

        tempdir = tempfile.mkdtemp(prefix='delugeweb-')
        tmp_file = os.path.join(tempdir, url.split('/')[-1])
        log.debug('filename: %s', tmp_file)
        headers = {}
        if cookie:
            headers['Cookie'] = cookie
            log.debug('cookie: %s', cookie)
        d = httpdownloader.download_file(url, tmp_file, headers=headers)
        d.addCallbacks(on_download_success, on_download_fail)
        return d

    @export
    def get_torrent_info(self, filename):
        """
        Return information about a torrent on the filesystem.

        :param filename: the path to the torrent
        :type filename: string

        :returns: information about the torrent:

        ::

            {
                "name": the torrent name,
                "files_tree": the files the torrent contains,
                "info_hash" the torrents info_hash
            }

        :rtype: dictionary
        """
        try:
            torrent_info = TorrentInfo(filename.strip(), 2)
            return torrent_info.as_dict('name', 'info_hash', 'files_tree')
        except Exception as ex:
            log.error(ex)
            return False

    @export
    def get_magnet_info(self, uri):
        """Parse a magnet URI for hash and name."""
        return get_magnet_info(uri)

    @export
    def add_torrents(self, torrents):
        """
        Add torrents by file

        :param torrents: A list of dictionaries containing the torrent \
            path and torrent options to add with.
        :type torrents: list

        ::

            json_api.web.add_torrents([{
                "path": "/tmp/deluge-web/some-torrent-file.torrent",
                "options": {"download_location": "/home/deluge/"}
            }])

        """
        deferreds = []

        for torrent in torrents:
            if is_magnet(torrent['path']):
                log.info(
                    'Adding torrent from magnet uri `%s` with options `%r`',
                    torrent['path'],
                    torrent['options'],
                )
                d = client.core.add_torrent_magnet(torrent['path'], torrent['options'])
                deferreds.append(d)
            else:
                filename = os.path.basename(torrent['path'])
                with open(torrent['path'], 'rb') as _file:
                    fdump = b64encode(_file.read())
                log.info(
                    'Adding torrent from file `%s` with options `%r`',
                    filename,
                    torrent['options'],
                )
                d = client.core.add_torrent_file_async(
                    filename, fdump, torrent['options']
                )
                deferreds.append(d)
        return DeferredList(deferreds, consumeErrors=False)

    @export
    def get_hosts(self):
        """
        Return the hosts in the hostlist.
        """
        log.debug('get_hosts called')
        return self.hostlist.get_hosts_info()

    @export
    def get_host_status(self, host_id):
        """
        Returns the current status for the specified host.

        :param host_id: the hash id of the host
        :type host_id: string

        """

        def response(result):
            return result

        return self.hostlist.get_host_status(host_id).addCallback(response)

    @export
    def add_host(self, host, port, username='', password=''):
        """Adds a host to the list.

        Args:
            host (str): The IP or hostname of the deluge daemon.
            port (int): The port of the deluge daemon.
            username (str): The username to login to the daemon with.
            password (str): The password to login to the daemon with.

        Returns:
            tuple: A tuple of (bool, str). If True will contain the host_id, otherwise
                if False will contain the error message.
        """
        try:
            host_id = self.hostlist.add_host(host, port, username, password)
        except ValueError as ex:
            return False, str(ex)
        else:
            return True, host_id

    @export
    def edit_host(self, host_id, host, port, username='', password=''):
        """Edit host details in the hostlist.

        Args:
            host_id (str): The host identifying hash.
            host (str): The IP or hostname of the deluge daemon.
            port (int): The port of the deluge daemon.
            username (str): The username to login to the daemon with.
            password (str): The password to login to the daemon with.

        Returns:
            bool: True if successful, False otherwise.

        """
        return self.hostlist.update_host(host_id, host, port, username, password)

    @export
    def remove_host(self, host_id):
        """Removes a host from the hostlist.

        Args:
            host_id (str): The host identifying hash.

        Returns:
            bool: True if successful, False otherwise.

        """
        return self.hostlist.remove_host(host_id)

    @export
    def start_daemon(self, port):
        """
        Starts a local daemon.
        """
        client.start_daemon(port, get_config_dir())

    @export
    @maybe_coroutine
    async def stop_daemon(self, host_id):
        try:
            await self.hostlist.connect_host(host_id)
        except Exception as err:
            msg = f'Error occurred stopping daemon: {err}'
            result = (False, msg)
        else:
            client.daemon.shutdown()
            result = (True,)

        return Deferred().callback(result)

    @export
    def get_config(self):
        """
        Get the configuration dictionary for the web interface.

        :rtype: dictionary
        :returns: the configuration
        """
        config = component.get('DelugeWeb').config.config.copy()
        del config['sessions']
        del config['pwd_salt']
        del config['pwd_sha1']
        return config

    @export
    def set_config(self, config):
        """
        Sets the configuration dictionary for the web interface.

        :param config: The configuration options to update
        :type config: dictionary
        """
        web_config = component.get('DelugeWeb').config
        for key in config:
            if key in ['sessions', 'pwd_salt', 'pwd_sha1']:
                log.warning('Ignored attempt to overwrite web config key: %s', key)
                continue
            web_config[key] = config[key]

    @export
    def get_plugins(self):
        """All available and enabled plugins within WebUI.

        Note:
            This does not represent all plugins from deluge.client.core.

        Returns:
            dict: A dict containing 'available_plugins' and 'enabled_plugins' lists.

        """

        return {
            'enabled_plugins': list(component.get('Web.PluginManager').plugins),
            'available_plugins': component.get('Web.PluginManager').available_plugins,
        }

    @export
    def get_plugin_info(self, name):
        """Get the details for a plugin."""
        return component.get('Web.PluginManager').get_plugin_info(name)

    @export
    def get_plugin_resources(self, name):
        """Get the resource data files for a plugin."""
        return component.get('Web.PluginManager').get_plugin_resources(name)

    @export
    def upload_plugin(self, filename, path):
        """Upload a plugin to config."""
        main_deferred = Deferred()

        shutil.copyfile(path, os.path.join(get_config_dir(), 'plugins', filename))
        component.get('Web.PluginManager').scan_for_plugins()

        if client.is_localhost():
            client.core.rescan_plugins()
            return True
        with open(path, 'rb') as _file:
            plugin_data = b64encode(_file.read())

        def on_upload_complete(*args):
            client.core.rescan_plugins()
            component.get('Web.PluginManager').scan_for_plugins()
            main_deferred.callback(True)

        def on_upload_error(*args):
            main_deferred.callback(False)

        d = client.core.upload_plugin(filename, plugin_data)
        d.addCallback(on_upload_complete)
        d.addErrback(on_upload_error)
        return main_deferred

    @export
    def register_event_listener(self, event):
        """
        Add a listener to the event queue.

        :param event: The event name
        :type event: string
        """
        self.event_queue.add_listener(__request__.session_id, event)

    @export
    def deregister_event_listener(self, event):
        """
        Remove an event listener from the event queue.

        :param event: The event name
        :type event: string
        """
        self.event_queue.remove_listener(__request__.session_id, event)

    @export
    def get_events(self):
        """
        Retrieve the pending events for the session.
        """
        return self.event_queue.get_events(__request__.session_id)

    @export
    def set_theme(self, theme):
        """
        Sets a new Theme to the WebUI

        Args:
            theme (str): the theme to apply
        """
        component.get('DelugeWeb').set_theme(theme)


class WebUtils(JSONComponent):
    """
    Utility functions for the Web UI that do not fit in the WebApi.
    """

    def __init__(self):
        super().__init__('WebUtils')

    @export
    def get_languages(self):
        """
        Get the available translated languages

        Returns:
             list: of tuples ``[(lang-id, language-name), ...]``
        """
        return get_languages()

    @export
    def get_themes(self):
        """
        Get the available themes

        Returns:
            list: of themes ``[theme1, theme2, ...]``
        """
        return component.get('DelugeWeb').get_themes()



#
# Copyright (C) 2008,2009 Andrew Resch <andrewresch@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

"""RPCServer Module"""

import logging
import os
import sys
import traceback
from collections import namedtuple
from types import FunctionType
from typing import Callable, TypeVar, overload

from twisted.internet import defer, reactor
from twisted.internet.protocol import Factory, connectionDone

import deluge.component as component
import deluge.configmanager
from deluge.core.authmanager import (
    AUTH_LEVEL_ADMIN,
    AUTH_LEVEL_DEFAULT,
    AUTH_LEVEL_NONE,
)
from deluge.crypto_utils import check_ssl_keys, get_context_factory
from deluge.error import (
    BadLoginError,
    DelugeError,
    IncompatibleClient,
    NotAuthorizedError,
    WrappedException,
    _ClientSideRecreateError,
)
from deluge.event import ClientDisconnectedEvent
from deluge.transfer import DelugeTransferProtocol

RPC_RESPONSE = 1
RPC_ERROR = 2
RPC_EVENT = 3

log = logging.getLogger(__name__)

TCallable = TypeVar('TCallable', bound=Callable)


@overload
def export(func: TCallable) -> TCallable: ...


@overload
def export(auth_level: int) -> Callable[[TCallable], TCallable]: ...


def export(auth_level=AUTH_LEVEL_DEFAULT):
    """
    Decorator function to register an object's method as an RPC.  The object
    will need to be registered with an :class:`RPCServer` to be effective.

    :param func: the function to export
    :type func: function
    :param auth_level: the auth level required to call this method
    :type auth_level: int

    """

    def wrap(func, *args, **kwargs):
        func._rpcserver_export = True
        func._rpcserver_auth_level = auth_level

        rpc_text = '**RPC exported method** (*Auth level: %s*)' % auth_level

        # Append the RPC text while ensuring correct docstring formatting.
        if func.__doc__:
            if func.__doc__.endswith('    '):
                indent = func.__doc__.split('\n')[-1]
                func.__doc__ += f'\n{indent}'
            else:
                func.__doc__ += '\n\n'
            func.__doc__ += rpc_text
        else:
            func.__doc__ = rpc_text

        return func

    if isinstance(auth_level, FunctionType):
        func = auth_level
        auth_level = AUTH_LEVEL_DEFAULT
        return wrap(func)
    else:
        return wrap


def format_request(call):
    """
    Format the RPCRequest message for debug printing

    :param call: the request
    :type call: a RPCRequest

    :returns: a formatted string for printing
    :rtype: str

    """
    try:
        s = call[1] + '('
        if call[2]:
            s += ', '.join([str(x) for x in call[2]])
        if call[3]:
            if call[2]:
                s += ', '
            s += ', '.join([key + '=' + str(value) for key, value in call[3].items()])
        s += ')'
    except UnicodeEncodeError:
        return 'UnicodeEncodeError, call: %s' % call
    else:
        return s


class DelugeRPCProtocol(DelugeTransferProtocol):
    def __init__(self):
        super().__init__()
        # namedtuple subclass with auth_level, username for the connected session.
        self.AuthLevel = namedtuple('SessionAuthlevel', 'auth_level, username')

    def message_received(self, request):
        """
        This method is called whenever a message is received from a client.  The
        only message that a client sends to the server is a RPC Request message.
        If the RPC Request message is valid, then the method is called in
        :meth:`dispatch`.

        :param request: the request from the client.
        :type data: tuple

        """
        if not isinstance(request, tuple):
            log.debug('Received invalid message: type is not tuple')
            return

        if len(request) < 1:
            log.debug('Received invalid message: there are no items')
            return

        for call in request:
            if len(call) != 4:
                log.debug(
                    'Received invalid rpc request: number of items ' 'in request is %s',
                    len(call),
                )
                continue
            # log.debug('RPCRequest: %s', format_request(call))
            reactor.callLater(0, self.dispatch, *call)

    def sendData(self, data):  # NOQA: N802
        """
        Sends the data to the client.

        :param data: the object that is to be sent to the client.  This should
            be one of the RPC message types.
        :type data: object

        """
        try:
            self.transfer_message(data)
        except Exception as ex:
            log.warning('Error occurred when sending message: %s.', ex)
            log.exception(ex)
            raise

    def connectionMade(self):  # NOQA: N802
        """
        This method is called when a new client connects.
        """
        peer = self.transport.getPeer()
        log.info('Deluge Client connection made from: %s:%s', peer.host, peer.port)
        # Set the initial auth level of this session to AUTH_LEVEL_NONE
        self.factory.authorized_sessions[self.transport.sessionno] = self.AuthLevel(
            AUTH_LEVEL_NONE, ''
        )

    def connectionLost(self, reason=connectionDone):  # NOQA: N802
        """
        This method is called when the client is disconnected.

        :param reason: the reason the client disconnected.
        :type reason: str

        """

        # We need to remove this session from various dicts
        del self.factory.authorized_sessions[self.transport.sessionno]
        if self.transport.sessionno in self.factory.session_protocols:
            del self.factory.session_protocols[self.transport.sessionno]
        if self.transport.sessionno in self.factory.interested_events:
            del self.factory.interested_events[self.transport.sessionno]

        if self.factory.state == 'running':
            component.get('EventManager').emit(
                ClientDisconnectedEvent(self.factory.session_id)
            )
        log.info('Deluge client disconnected: %s', reason.value)

    def valid_session(self):
        return self.transport.sessionno in self.factory.authorized_sessions

    def dispatch(self, request_id, method, args, kwargs):
        """
        This method is run when a RPC Request is made.  It will run the local method
        and will send either a RPC Response or RPC Error back to the client.

        :param request_id: the request_id from the client (sent in the RPC Request)
        :type request_id: int
        :param method: the local method to call. It must be registered with
            the :class:`RPCServer`.
        :type method: str
        :param args: the arguments to pass to `method`
        :type args: list
        :param kwargs: the keyword-arguments to pass to `method`
        :type kwargs: dict

        """

        def send_error():
            """
            Sends an error response with the contents of the exception that was raised.
            """
            exc_type, exc_value, dummy_exc_trace = sys.exc_info()
            formated_tb = traceback.format_exc()
            try:
                self.sendData(
                    (
                        RPC_ERROR,
                        request_id,
                        exc_type.__name__,
                        exc_value._args,
                        exc_value._kwargs,
                        formated_tb,
                    )
                )
            except AttributeError:
                # This is not a deluge exception (object has no attribute '_args), let's wrap it
                log.warning(
                    'An exception occurred while sending RPC_ERROR to '
                    'client. Wrapping it and resending. Error to '
                    'send(causing exception goes next):\n%s',
                    formated_tb,
                )
                try:
                    raise WrappedException(
                        str(exc_value), exc_type.__name__, formated_tb
                    )
                except WrappedException:
                    send_error()
            except Exception as ex:
                log.error(
                    'An exception occurred while sending RPC_ERROR to client: %s', ex
                )

        if method == 'daemon.info':
            # This is a special case and used in the initial connection process
            self.sendData((RPC_RESPONSE, request_id, deluge.common.get_version()))
            return
        elif method == 'daemon.login':
            # This is a special case and used in the initial connection process
            # We need to authenticate the user here
            log.debug('RPC dispatch daemon.login')
            try:
                client_version = kwargs.pop('client_version', None)
                if client_version is None:
                    raise IncompatibleClient(deluge.common.get_version())
                ret = component.get('AuthManager').authorize(*args, **kwargs)
                if ret:
                    self.factory.authorized_sessions[self.transport.sessionno] = (
                        self.AuthLevel(ret, args[0])
                    )
                    self.factory.session_protocols[self.transport.sessionno] = self
            except Exception as ex:
                send_error()
                if not isinstance(ex, _ClientSideRecreateError):
                    log.exception(ex)
                if isinstance(ex, BadLoginError):
                    peer = self.transport.getPeer()
                    log.error(
                        'Deluge client authentication error made from: %s:%s (%s)',
                        peer.host,
                        peer.port,
                        str(ex),
                    )
            else:
                self.sendData((RPC_RESPONSE, request_id, (ret)))
                if not ret:
                    self.transport.loseConnection()
            return

        # Anything below requires a valid session
        if not self.valid_session():
            return

        if method == 'daemon.set_event_interest':
            log.debug('RPC dispatch daemon.set_event_interest')
            # This special case is to allow clients to set which events they are
            # interested in receiving.
            # We are expecting a sequence from the client.
            try:
                if self.transport.sessionno not in self.factory.interested_events:
                    self.factory.interested_events[self.transport.sessionno] = []
                self.factory.interested_events[self.transport.sessionno].extend(args[0])
            except Exception:
                send_error()
            else:
                self.sendData((RPC_RESPONSE, request_id, (True)))
            return

        if method not in self.factory.methods:
            try:
                # Raise exception to be sent back to client
                raise AttributeError('RPC call on invalid function: %s' % method)
            except AttributeError:
                send_error()
                return

        log.debug('RPC dispatch %s', method)
        try:
            method_auth_requirement = self.factory.methods[method]._rpcserver_auth_level
            auth_level = self.factory.authorized_sessions[
                self.transport.sessionno
            ].auth_level
            if auth_level < method_auth_requirement:
                # This session is not allowed to call this method
                log.debug(
                    'Session %s is attempting an unauthorized method call!',
                    self.transport.sessionno,
                )
                raise NotAuthorizedError(auth_level, method_auth_requirement)
            # Set the session_id in the factory so that methods can know
            # which session is calling it.
            self.factory.session_id = self.transport.sessionno
            ret = self.factory.methods[method](*args, **kwargs)
        except Exception as ex:
            send_error()
            # Don't bother printing out DelugeErrors, because they are just
            # for the client
            if not isinstance(ex, DelugeError):
                log.exception('Exception calling RPC request: %s', ex)
        else:
            # Check if the return value is a deferred, since we'll need to
            # wait for it to fire before sending the RPC_RESPONSE
            if isinstance(ret, defer.Deferred):

                def on_success(result):
                    try:
                        self.sendData((RPC_RESPONSE, request_id, result))
                    except Exception:
                        send_error()
                    return result

                def on_fail(failure):
                    try:
                        failure.raiseException()
                    except Exception:
                        send_error()
                    return failure

                ret.addCallbacks(on_success, on_fail)
            else:
                self.sendData((RPC_RESPONSE, request_id, ret))


class RPCServer(component.Component):
    """
    This class is used to handle rpc requests from the client.  Objects are
    registered with this class and their methods are exported using the export
    decorator.

    :param port: the port the RPCServer will listen on
    :type port: int
    :param interface: the interface to listen on, this may override the `allow_remote` setting
    :type interface: str
    :param allow_remote: set True if the server should allow remote connections
    :type allow_remote: bool
    :param listen: if False, will not start listening.. This is only useful in Classic Mode
    :type listen: bool
    """

    def __init__(self, port=58846, interface='', allow_remote=False, listen=True):
        component.Component.__init__(self, 'RPCServer')

        self.factory = Factory()
        self.factory.protocol = DelugeRPCProtocol
        self.factory.session_id = -1
        self.factory.state = 'running'

        # Holds the registered methods
        self.factory.methods = {}
        # Holds the session_ids and auth levels
        self.factory.authorized_sessions = {}
        # Holds the protocol objects with the session_id as key
        self.factory.session_protocols = {}
        # Holds the interested event list for the sessions
        self.factory.interested_events = {}

        self.listen = listen
        if not listen:
            return

        if allow_remote:
            hostname = ''
        else:
            hostname = 'localhost'

        if interface:
            hostname = interface

        log.info('Starting DelugeRPC server %s:%s', hostname, port)

        # Check for SSL keys and generate some if needed
        check_ssl_keys()

        cert = os.path.join(deluge.configmanager.get_config_dir('ssl'), 'daemon.cert')
        pkey = os.path.join(deluge.configmanager.get_config_dir('ssl'), 'daemon.pkey')

        try:
            reactor.listenSSL(
                port, self.factory, get_context_factory(cert, pkey), interface=hostname
            )
        except Exception as ex:
            log.debug('Daemon already running or port not available.: %s', ex)
            raise

    def register_object(self, obj, name=None):
        """
        Registers an object to export it's rpc methods.  These methods should
        be exported with the export decorator prior to registering the object.

        :param obj: the object that we want to export
        :type obj: object
        :param name: the name to use, if None, it will be the class name of the object
        :type name: str
        """
        if not name:
            name = obj.__class__.__name__.lower()

        for d in dir(obj):
            if d[0] == '_':
                continue
            if getattr(getattr(obj, d), '_rpcserver_export', False):
                log.debug('Registering method: %s', name + '.' + d)
                self.factory.methods[name + '.' + d] = getattr(obj, d)

    def deregister_object(self, obj):
        """
        Deregisters an objects exported rpc methods.

        :param obj: the object that was previously registered

        """
        for key, value in self.factory.methods.items():
            if value.__self__ == obj:
                del self.factory.methods[key]

    def get_object_method(self, name):
        """
        Returns a registered method.

        :param name: the name of the method, usually in the form of 'object.method'
        :type name: str

        :returns: method

        :raises KeyError: if `name` is not registered

        """
        return self.factory.methods[name]

    def get_method_list(self):
        """
        Returns a list of the exported methods.

        :returns: the exported methods
        :rtype: list
        """
        return list(self.factory.methods)

    def get_session_id(self):
        """
        Returns the session id of the current RPC.

        :returns: the session id, this will be -1 if no connections have been made
        :rtype: int

        """
        return self.factory.session_id

    def get_session_user(self):
        """
        Returns the username calling the current RPC.

        :returns: the username of the user calling the current RPC
        :rtype: string

        """
        if not self.listen:
            return 'localclient'
        session_id = self.get_session_id()
        if session_id > -1 and session_id in self.factory.authorized_sessions:
            return self.factory.authorized_sessions[session_id].username
        else:
            # No connections made yet
            return ''

    def get_session_auth_level(self):
        """
        Returns the auth level of the user calling the current RPC.

        :returns: the auth level
        :rtype: int
        """
        if not self.listen or not self.is_session_valid(self.get_session_id()):
            return AUTH_LEVEL_ADMIN
        return self.factory.authorized_sessions[self.get_session_id()].auth_level

    def get_rpc_auth_level(self, rpc):
        """
        Returns the auth level requirement for an exported rpc.

        :returns: the auth level
        :rtype: int
        """
        return self.factory.methods[rpc]._rpcserver_auth_level

    def is_session_valid(self, session_id):
        """
        Checks if the session is still valid, eg, if the client is still connected.

        :param session_id: the session id
        :type session_id: int

        :returns: True if the session is valid
        :rtype: bool

        """
        return session_id in self.factory.authorized_sessions

    def emit_event(self, event):
        """
        Emits the event to interested clients.

        :param event: the event to emit
        :type event: :class:`deluge.event.DelugeEvent`
        """
        log.debug('intevents: %s', self.factory.interested_events)
        # Use copy of `interested_events` since it can mutate while iterating.
        for session_id, interest in self.factory.interested_events.copy().items():
            if event.name in interest:
                log.debug('Emit Event: %s %s', event.name, event.args)
                # This session is interested so send a RPC_EVENT
                self.factory.session_protocols[session_id].sendData(
                    (RPC_EVENT, event.name, event.args)
                )

    def emit_event_for_session_id(self, session_id, event):
        """
        Emits the event to specified session_id.

        :param session_id: the event to emit
        :type session_id: int
        :param event: the event to emit
        :type event: :class:`deluge.event.DelugeEvent`
        """
        if not self.is_session_valid(session_id):
            log.debug(
                'Session ID %s is not valid. Not sending event "%s".',
                session_id,
                event.name,
            )
            return
        if session_id not in self.factory.interested_events:
            log.debug(
                'Session ID %s is not interested in any events. Not sending event "%s".',
                session_id,
                event.name,
            )
            return
        if event.name not in self.factory.interested_events[session_id]:
            log.debug(
                'Session ID %s is not interested in event "%s". Not sending it.',
                session_id,
                event.name,
            )
            return
        log.debug(
            'Sending event "%s" with args "%s" to session id "%s".',
            event.name,
            event.args,
            session_id,
        )
        self.factory.session_protocols[session_id].sendData(
            (RPC_EVENT, event.name, event.args)
        )

    def stop(self):
        self.factory.state = 'stopping'



#
# Copyright (C) 2008-2009 Ido Abramovich <ido.deluge@gmail.com>
# Copyright (C) 2009 Andrew Resch <andrewresch@gmail.com>
# Copyright (C) 2011 Nick Lanham <nick@afternight.org>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import glob
import logging
import os
from base64 import b64encode

import deluge.common
from deluge.ui.client import client
from deluge.ui.common import TorrentInfo

log = logging.getLogger(__name__)


def _bracket_fixup(path):
    if path.find('[') == -1 and path.find(']') == -1:
        return path
    sentinal = 256
    while path.find(chr(sentinal)) != -1:
        sentinal += 1
        if sentinal > 65535:
            log.error(
                'Cannot fix brackets in path, path contains all possible sentinal characters'
            )
            return path
    newpath = path.replace(']', chr(sentinal))
    newpath = newpath.replace('[', '[[]')
    newpath = newpath.replace(chr(sentinal), '[]]')
    return newpath


def add_torrent(t_file, options, success_cb, fail_cb, ress):
    t_options = {}
    if options['path']:
        t_options['download_location'] = os.path.expanduser(options['path'])
    t_options['add_paused'] = options['add_paused']

    is_url = (options['path_type'] != 1) and (
        deluge.common.is_url(t_file) or options['path_type'] == 2
    )
    is_magnet = (
        not (is_url) and (options['path_type'] != 1) and deluge.common.is_magnet(t_file)
    )

    if is_url or is_magnet:
        files = [t_file]
    else:
        files = glob.glob(_bracket_fixup(t_file))
    num_files = len(files)
    ress['total'] = num_files

    if num_files <= 0:
        fail_cb('Does not exist', t_file, ress)

    for f in files:
        if is_url:
            client.core.add_torrent_url(f, t_options).addCallback(
                success_cb, f, ress
            ).addErrback(fail_cb, f, ress)
        elif is_magnet:
            client.core.add_torrent_magnet(f, t_options).addCallback(
                success_cb, f, ress
            ).addErrback(fail_cb, f, ress)
        else:
            if not os.path.exists(f):
                fail_cb('Does not exist', f, ress)
                continue
            if not os.path.isfile(f):
                fail_cb('Is a directory', f, ress)
                continue

            try:
                TorrentInfo(f)
            except Exception as ex:
                fail_cb(ex.message, f, ress)
                continue

            filename = os.path.split(f)[-1]
            with open(f, 'rb') as _file:
                filedump = b64encode(_file.read())

            client.core.add_torrent_file_async(
                filename, filedump, t_options
            ).addCallback(success_cb, f, ress).addErrback(fail_cb, f, ress)



#
# Copyright (C) Calum Lind 2017 <calumlind+deluge@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import logging
import os
import uuid
from socket import gaierror, getaddrinfo

from twisted.internet import defer

from deluge.common import get_localhost_auth
from deluge.config import Config
from deluge.configmanager import get_config_dir
from deluge.ui.client import Client, client

log = logging.getLogger(__name__)

DEFAULT_HOST = '127.0.0.1'
DEFAULT_PORT = 58846
LOCALHOST = ('127.0.0.1', 'localhost', '::1')


def default_hostlist():
    """Create a new hosts key for hostlist with a localhost entry"""
    host_id = uuid.uuid4().hex
    username, password = get_localhost_auth()
    return {'hosts': [(host_id, DEFAULT_HOST, DEFAULT_PORT, username, password)]}


def validate_host_info(hostname, port):
    """Checks that hostname and port are valid.

    Args:
        hostname (str): The IP or hostname of the deluge daemon.
        port (int): The port of the deluge daemon.

    Raises:
        ValueError: Host details are not valid with reason.
    """

    try:
        getaddrinfo(hostname, None)
    except gaierror as ex:
        raise ValueError('Host %s: %s', hostname, ex.args[1])

    if not isinstance(port, int):
        raise ValueError('Invalid port. Must be an integer')
    if not 0 <= port <= 65535:
        raise ValueError('Invalid port. Must be between 0-65535')


def migrate_hostlist(old_filename, new_filename):
    """Check for old hostlist filename and save details to new filename"""
    old_hostlist = get_config_dir(old_filename)
    if os.path.isfile(old_hostlist):
        config_v2 = Config(old_filename, config_dir=get_config_dir())
        config_v2.save(get_config_dir(new_filename))
        del config_v2

        try:
            os.rename(old_hostlist, old_hostlist + '.old')
        except OSError as ex:
            log.exception(ex)

        try:
            os.remove(old_hostlist + '.bak')
        except OSError:
            pass


def migrate_config_2_to_3(config):
    """Mirgrates old hostlist config files to new file version"""
    localclient_username, localclient_password = get_localhost_auth()
    if not localclient_username:
        # Nothing to do here, there's no auth file
        return
    for idx, (__, host, __, username, __) in enumerate(config['hosts'][:]):
        if host in LOCALHOST and not username:
            config['hosts'][idx][3] = localclient_username
            config['hosts'][idx][4] = localclient_password
    return config


def mask_hosts_password(hosts):
    """Replace passwords in hosts list with *'s for log output"""
    if not hosts:
        return hosts

    return [list(host)[:-1] + ['*' * 10] for host in hosts]


class HostList:
    """This class contains methods for adding, removing and looking up hosts in hostlist.conf."""

    def __init__(self):
        migrate_hostlist('hostlist.conf.1.2', 'hostlist.conf')
        self.config = Config(
            'hostlist.conf',
            default_hostlist(),
            config_dir=get_config_dir(),
            file_version=3,
            log_mask_funcs={'hosts': mask_hosts_password},
        )
        self.config.run_converter((1, 2), 3, migrate_config_2_to_3)
        self.config.save()

    def check_info_exists(self, hostname, port, username, skip_host_id=None):
        """Check for existing host entries with the same details.

        Args:
            hostname (str): The IP or hostname of the deluge daemon.
            port (int): The port of the deluge daemon.
            username (str): The username to login to the daemon with.
            skip_host_id (str): A host_id to skip to check if other hosts match details.

        Raises:
            ValueError: Host details already exist.

        """
        for host_entry in self.config['hosts']:
            if (hostname, port, username) == (
                host_entry[1],
                host_entry[2],
                host_entry[3],
            ):
                if skip_host_id is not None and skip_host_id == host_entry[0]:
                    continue
                raise ValueError('Host details already in hostlist')

    def add_host(self, hostname, port, username, password):
        """Add a new host to hostlist.

        Args:
            hostname (str): The IP or hostname of the deluge daemon.
            port (int): The port of the deluge daemon.
            username (str): The username to login to the daemon with.
            password (str): The password to login to the daemon with.

        Returns:
            str: The new host id.
        """
        if (
            not password and not username or username == 'localclient'
        ) and hostname in LOCALHOST:
            username, password = get_localhost_auth()

        validate_host_info(hostname, port)
        self.check_info_exists(hostname, port, username)
        host_id = uuid.uuid4().hex
        self.config['hosts'].append((host_id, hostname, port, username, password))
        self.config.save()
        return host_id

    def get_host_info(self, host_id):
        """Get the host details for host_id.

        Args:
            host_id (str): The host id to get info on.

        Returns:
            list: A list of (host_id, hostname, port, username).

        """
        for host_entry in self.config['hosts']:
            if host_entry[0] == host_id:
                return host_entry[0:4]
        else:
            return []

    def get_hosts_info(self):
        """Get information of all the hosts in the hostlist.

        Returns:
            list of lists: Host information in the format [(host_id, hostname, port, username)].

        """
        return [host_entry[0:4] for host_entry in self.config['hosts']]

    def get_host_status(self, host_id):
        """Gets the current status (online/offline) of the host

        Args:
            host_id (str): The host id to check status of.

        Returns:
            tuple: A tuple of strings (host_id, status, version).

        """
        status_offline = (host_id, 'Offline', '')

        def on_connect(result, c, host_id):
            """Successfully connected to a daemon"""

            def on_info(info, c):
                c.disconnect()
                return host_id, 'Online', info

            def on_info_fail(reason, c):
                c.disconnect()
                return status_offline

            return c.daemon.info().addCallback(on_info, c).addErrback(on_info_fail, c)

        def on_connect_failed(reason, host_id):
            """Connection to daemon failed"""
            log.debug('Host status failed for %s: %s', host_id, reason)
            return status_offline

        try:
            host_id, host, port, user = self.get_host_info(host_id)
        except ValueError:
            log.warning('Problem getting host_id info from hostlist')
            return defer.succeed(status_offline)

        try:
            ips = list({addrinfo[4][0] for addrinfo in getaddrinfo(host, None)})
        except (gaierror, IndexError) as ex:
            log.warning('Unable to resolve host %s to IP: %s', host, ex.args[1])
            return defer.succeed(status_offline)

        host_conn_list = [
            (
                host_ip,
                port,
                'localclient' if not user and host_ip in LOCALHOST else user,
            )
            for host_ip in ips
        ]

        for host_conn_info in host_conn_list:
            if client.connected() and host_conn_info == client.connection_info():
                # Currently connected to host_id daemon.
                def on_info(info, host_id):
                    log.debug('Client connected, query info: %s', info)
                    return host_id, 'Connected', info

                return client.daemon.info().addCallback(on_info, host_id)
            else:
                # Attempt to connect to daemon with host_id details.
                c = Client()
                d = c.connect(host, port, skip_authentication=True)
                d.addCallback(on_connect, c, host_id)
                d.addErrback(on_connect_failed, host_id)
                return d

    def update_host(self, host_id, hostname, port, username, password):
        """Update the supplied host id with new connection details.

        Args:
            host_id (str): The host id to update.
            hostname (str): The new IP or hostname of the deluge daemon.
            port (int): The new port of the deluge daemon.
            username (str): The new username to login to the daemon with.
            password (str): The new password to login to the daemon with.

        """
        validate_host_info(hostname, port)
        self.check_info_exists(hostname, port, username, skip_host_id=host_id)

        if (
            not password and not username or username == 'localclient'
        ) and hostname in LOCALHOST:
            username, password = get_localhost_auth()

        for idx, host_entry in enumerate(self.config['hosts']):
            if host_id == host_entry[0]:
                self.config['hosts'][idx] = host_id, hostname, port, username, password
                self.config.save()
                return True
        return False

    def remove_host(self, host_id):
        """Removes the host entry from hostlist config.

        Args:
            host_id (str): The host id to remove.

        Returns:
            bool: True is successfully removed, False otherwise.

        """
        for host_entry in self.config['hosts']:
            if host_id == host_entry[0]:
                self.config['hosts'].remove(host_entry)
                self.config.save()
                return True
        else:
            return False

    def add_default_host(self):
        self.add_host(DEFAULT_HOST, DEFAULT_PORT, *get_localhost_auth())

    def connect_host(self, host_id):
        """Connect to host daemon"""
        for host_entry in self.config['hosts']:
            if host_entry[0] == host_id:
                __, host, port, username, password = host_entry
                return client.connect(host, port, username, password)

        return defer.fail(Exception('Bad host id'))



#
# Copyright (C) 2007,2008 Andrew Resch <andrewresch@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import os
import stat
from datetime import datetime, timedelta, timezone

from cryptography import x509
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.x509.oid import NameOID
from OpenSSL.crypto import FILETYPE_PEM
from twisted.internet.ssl import (
    AcceptableCiphers,
    Certificate,
    CertificateOptions,
    KeyPair,
    TLSVersion,
)

import deluge.configmanager

# A TLS ciphers list.
# Sources for more information on TLS ciphers:
# - https://wiki.mozilla.org/Security/Server_Side_TLS
# - https://www.ssllabs.com/projects/best-practices/index.html
# - https://hynek.me/articles/hardening-your-web-servers-ssl-ciphers/
#
# This list was inspired by the `urllib3` library
# - https://github.com/urllib3/urllib3/blob/master/urllib3/util/ssl_.py#L79
#
# The general intent is:
# - prefer cipher suites that offer perfect forward secrecy (ECDHE),
# - prefer AES-GCM over ChaCha20 because hardware-accelerated AES is common,
# - disable NULL authentication, MD5 MACs and DSS for security reasons.
TLS_CIPHERS = ':'.join(
    [
        'ECDH+AESGCM',
        'ECDH+CHACHA20',
        'AES256-GCM-SHA384',
        'AES128-GCM-SHA256',
        '!DSS' '!aNULL',
        '!eNULL',
        '!MD5',
    ]
)

# This value tells OpenSSL to disable all SSL/TLS renegotiation.
SSL_OP_NO_RENEGOTIATION = 0x40000000


def get_context_factory(cert_path, pkey_path):
    """OpenSSL context factory.

    Generates an OpenSSL context factory using Twisted's CertificateOptions class.
    This will keep a server cipher order.

    Args:
        cert_path (string): The path to the certificate file
        pkey_path (string): The path to the private key file

    Returns:
        twisted.internet.ssl.CertificateOptions: An OpenSSL context factory
    """

    with open(cert_path) as cert:
        certificate = Certificate.loadPEM(cert.read()).original
    with open(pkey_path) as pkey:
        private_key = KeyPair.load(pkey.read(), FILETYPE_PEM).original
    ciphers = AcceptableCiphers.fromOpenSSLCipherString(TLS_CIPHERS)
    cert_options = CertificateOptions(
        privateKey=private_key,
        certificate=certificate,
        raiseMinimumTo=TLSVersion.TLSv1_2,
        acceptableCiphers=ciphers,
    )
    ctx = cert_options.getContext()
    ctx.use_certificate_chain_file(cert_path)
    ctx.set_options(SSL_OP_NO_RENEGOTIATION)

    return cert_options


def check_ssl_keys():
    """
    Check for SSL cert/key and create them if necessary
    """
    ssl_dir = deluge.configmanager.get_config_dir('ssl')
    if not os.path.exists(ssl_dir):
        # The ssl folder doesn't exist so we need to create it
        os.makedirs(ssl_dir)
        generate_ssl_keys()
    else:
        for f in ('daemon.pkey', 'daemon.cert'):
            if not os.path.exists(os.path.join(ssl_dir, f)):
                generate_ssl_keys()
                break


def generate_ssl_keys():
    """
    This method generates a new SSL key/cert.
    """
    # Generate key pair
    pkey = rsa.generate_private_key(public_exponent=65537, key_size=2048)

    # Generate certificate
    subject = x509.Name([x509.NameAttribute(NameOID.COMMON_NAME, 'Deluge Daemon')])
    now = datetime.now(timezone.utc)
    cert = (
        x509.CertificateBuilder()
        .subject_name(subject)
        .issuer_name(subject)
        .public_key(pkey.public_key())
        .serial_number(x509.random_serial_number())
        .not_valid_before(now)
        .not_valid_after(now + timedelta(days=365 * 3))  # Three Years
        .sign(pkey, hashes.SHA256())
    )

    ssl_dir = deluge.configmanager.get_config_dir('ssl')
    files = {
        'daemon.pkey': pkey.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.PKCS8,
            encryption_algorithm=serialization.NoEncryption(),
        ),
        'daemon.cert': cert.public_bytes(serialization.Encoding.PEM),
    }
    for name, data in files.items():
        path = os.path.join(ssl_dir, name)
        fd = os.open(path, os.O_WRONLY | os.O_CREAT | os.O_TRUNC, stat.S_IRUSR | stat.S_IWUSR)
        with os.fdopen(fd, 'wb') as _file:
            _file.write(data)
        os.chmod(path, stat.S_IRUSR | stat.S_IWUSR)



# SPDX-License-Identifier: GPL-3.0-or-later WITH GPL-3.0-linking-exception
from importlib import resources
from pathlib import Path

from deluge.pluginmanagerbase import _extract_egg


def test_extract_egg_creates_real_files(fake_egg, tmp_path):
    """Verify extraction produces real files on disk.

    Zip eggs serve resources from inside the archive, so non-Python assets
    (JS, UI files, etc.) must be extracted to disk to be accessible as real
    filesystem paths.
    """
    _, egg_path = fake_egg
    dest = tmp_path / 'extracted'
    _extract_egg(egg_path, dest)

    assert (dest / 'fake_plugin' / '__init__.py').exists()
    assert (dest / 'fake_plugin' / 'data' / 'script.js').exists()
    assert (
        dest / 'fake_plugin' / 'data' / 'script.js'
    ).read_text() == 'console.log("test");'


def test_extract_egg_skips_egg_info(fake_egg, tmp_path):
    """Test that EGG-INFO (setuptools package metadata) is excluded from extraction.

    Only importable code belongs in the cache dir, a top-level EGG-INFO directory
    could interfere with namespace package discovery by importlib.
    """
    _, egg_path = fake_egg
    dest = tmp_path / 'extracted'
    _extract_egg(egg_path, dest)

    assert not (dest / 'EGG-INFO').exists()


def test_plugin_dir_finder_resources_are_real_paths(installed_fake_plugin):
    """After activation via _PluginDirFinder, files() returns a real pathlib.Path."""
    resource = resources.files('fake_plugin') / 'data' / 'script.js'

    assert isinstance(resource, Path), f'Expected pathlib.Path, got {type(resource)}'
    assert resource.exists()
    assert resource.read_text() == 'console.log("test");'


def test_str_get_resource_gives_real_path(installed_fake_plugin):
    """Casting a resource path to str yields a real, resolvable filesystem path."""
    path_str = str(resources.files('fake_plugin') / 'data' / 'script.js')

    assert Path(path_str).exists(), f'Path does not exist: {path_str}'



#
# Copyright (C) 2012 Arek Stefański <asmageddon@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import logging
import os
from base64 import b64encode

import deluge.common
import deluge.component as component
from deluge.decorators import overrides
from deluge.ui.client import client
from deluge.ui.console.modes.basemode import BaseMode
from deluge.ui.console.modes.torrentlist.add_torrents_popup import report_add_status
from deluge.ui.console.utils import curses_util as util
from deluge.ui.console.utils import format_utils
from deluge.ui.console.widgets.popup import InputPopup, MessagePopup

try:
    import curses
except ImportError:
    pass

log = logging.getLogger(__name__)

# Big help string that gets displayed when the user hits 'h'
HELP_STR = """\
This screen allows you to browse and add torrent files located on your \
hard disk. Currently selected file is highlighted with a white background.
You can change the selected file using the up/down arrows or the \
PgUp/PgDown keys.  Home and End keys go to the first and last file \
in current directory respectively.

Select files with the 'm' key. Use 'M' for multi-selection. Press \
enter key to add them to session.

{!info!}'h'{!normal!} - Show this help

{!info!}'<'{!normal!} and {!info!}'>'{!normal!} - Change sort column and/or order

{!info!}'m'{!normal!} - Mark or unmark currently highlighted file
{!info!}'M'{!normal!} - Mark all files between current file and last selection.
{!info!}'c'{!normal!} - Clear selection.

{!info!}Left Arrow{!normal!} - Go up in directory hierarchy.
{!info!}Right Arrow{!normal!} - Enter currently highlighted folder.

{!info!}Enter{!normal!} - Enter currently highlighted folder or add torrents \
if a file is highlighted

{!info!}'q'{!normal!} - Go back to torrent overview
"""


class AddTorrents(BaseMode):
    def __init__(self, parent_mode, stdscr, console_config, encoding=None):
        self.console_config = console_config
        self.parent_mode = parent_mode
        self.popup = None
        self.view_offset = 0
        self.cursel = 0
        self.marked = set()
        self.last_mark = -1

        path = os.path.expanduser(self.console_config['addtorrents']['last_path'])

        self.path_stack = ['/'] + path.strip('/').split('/')
        self.path_stack_pos = len(self.path_stack)
        self.listing_files = []
        self.listing_dirs = []

        self.raw_rows = []
        self.raw_rows_files = []
        self.raw_rows_dirs = []
        self.formatted_rows = []

        self.sort_column = self.console_config['addtorrents']['sort_column']
        self.reverse_sort = self.console_config['addtorrents']['reverse_sort']

        BaseMode.__init__(self, stdscr, encoding)

        self._listing_space = self.rows - 5
        self.__refresh_listing()

        util.safe_curs_set(util.Curser.INVISIBLE)
        self.stdscr.notimeout(0)

    @overrides(component.Component)
    def start(self):
        pass

    @overrides(component.Component)
    def update(self):
        pass

    def __refresh_listing(self):
        path = os.path.join(*self.path_stack[: self.path_stack_pos])

        listing = os.listdir(path)

        self.listing_files = []
        self.listing_dirs = []

        self.raw_rows = []
        self.raw_rows_files = []
        self.raw_rows_dirs = []
        self.formatted_rows = []

        for f in listing:
            if os.path.isdir(os.path.join(path, f)):
                if self.console_config['addtorrents']['show_hidden_folders']:
                    self.listing_dirs.append(f)
                elif f[0] != '.':
                    self.listing_dirs.append(f)
            elif os.path.isfile(os.path.join(path, f)):
                if self.console_config['addtorrents']['show_misc_files']:
                    self.listing_files.append(f)
                elif f.endswith('.torrent'):
                    self.listing_files.append(f)

        for dirname in self.listing_dirs:
            row = []
            full_path = os.path.join(path, dirname)
            try:
                size = len(os.listdir(full_path))
            except OSError:
                size = -1
            time = os.stat(full_path).st_mtime

            row = [dirname, size, time, full_path, 1]

            self.raw_rows.append(row)
            self.raw_rows_dirs.append(row)

        # Highlight the directory we came from
        if self.path_stack_pos < len(self.path_stack):
            selected = self.path_stack[self.path_stack_pos]
            ld = sorted(self.listing_dirs, key=lambda n: n.lower())
            c = ld.index(selected)
            self.cursel = c

            if (self.view_offset + self._listing_space) <= self.cursel:
                self.view_offset = self.cursel - self._listing_space

        for filename in self.listing_files:
            row = []
            full_path = os.path.join(path, filename)
            size = os.stat(full_path).st_size
            time = os.stat(full_path).st_mtime

            row = [filename, size, time, full_path, 0]

            self.raw_rows.append(row)
            self.raw_rows_files.append(row)

        self.__sort_rows()

    def __sort_rows(self):
        self.console_config['addtorrents']['sort_column'] = self.sort_column
        self.console_config['addtorrents']['reverse_sort'] = self.reverse_sort
        self.console_config.save()

        self.raw_rows_dirs.sort(key=lambda r: r[0].lower())

        if self.sort_column == 'name':
            self.raw_rows_files.sort(
                key=lambda r: r[0].lower(), reverse=self.reverse_sort
            )
        elif self.sort_column == 'date':
            self.raw_rows_files.sort(key=lambda r: r[2], reverse=self.reverse_sort)
        self.raw_rows = self.raw_rows_dirs + self.raw_rows_files
        self.__refresh_rows()

    def __refresh_rows(self):
        self.formatted_rows = []

        for row in self.raw_rows:
            filename = deluge.common.decode_bytes(row[0])
            size = row[1]
            time = row[2]

            if row[4]:
                if size != -1:
                    size_str = '%i items' % size
                else:
                    size_str = ' unknown'

                cols = [filename, size_str, deluge.common.fdate(time)]
                widths = [self.cols - 35, 12, 23]
                self.formatted_rows.append(format_utils.format_row(cols, widths))
            else:
                # Size of .torrent file itself couldn't matter less so we'll leave it out
                cols = [filename, deluge.common.fdate(time)]
                widths = [self.cols - 23, 23]
                self.formatted_rows.append(format_utils.format_row(cols, widths))

    def scroll_list_up(self, distance):
        self.cursel -= distance
        if self.cursel < 0:
            self.cursel = 0

        if self.cursel < self.view_offset + 1:
            self.view_offset = max(self.cursel - 1, 0)

    def scroll_list_down(self, distance):
        self.cursel += distance
        if self.cursel >= len(self.formatted_rows):
            self.cursel = len(self.formatted_rows) - 1

        if (self.view_offset + self._listing_space) <= self.cursel + 1:
            self.view_offset = self.cursel - self._listing_space + 1

    def set_popup(self, pu):
        self.popup = pu
        self.refresh()

    @overrides(BaseMode)
    def on_resize(self, rows, cols):
        BaseMode.on_resize(self, rows, cols)
        if self.popup:
            self.popup.handle_resize()
        self._listing_space = self.rows - 5
        self.refresh()

    def refresh(self, lines=None):
        if self.mode_paused():
            return

        # Update the status bars
        self.stdscr.erase()
        self.draw_statusbars()

        off = 1

        # Render breadcrumbs
        s = 'Location: '
        for i, e in enumerate(self.path_stack):
            if e == '/':
                if i == self.path_stack_pos - 1:
                    s += '{!black,red,bold!}root'
                else:
                    s += '{!red,black,bold!}root'
            else:
                if i == self.path_stack_pos - 1:
                    s += '{!black,white,bold!}%s' % e
                else:
                    s += '{!white,black,bold!}%s' % e

            if e != len(self.path_stack) - 1:
                s += '{!white,black!}/'

        self.add_string(off, s)
        off += 1

        # Render header
        cols = ['Name', 'Contents', 'Modification time']
        widths = [self.cols - 35, 12, 23]
        s = ''
        for i, (c, w) in enumerate(zip(cols, widths)):
            cn = ''
            if i == 0:
                cn = 'name'
            elif i == 2:
                cn = 'date'

            if cn == self.sort_column:
                s += '{!black,green,bold!}' + c.ljust(w - 2)
                if self.reverse_sort:
                    s += '^ '
                else:
                    s += 'v '
            else:
                s += '{!green,black,bold!}' + c.ljust(w)
        self.add_string(off, s)
        off += 1

        # Render files and folders
        for i, row in enumerate(self.formatted_rows[self.view_offset :]):
            i += self.view_offset
            # It's a folder
            color_string = ''
            if self.raw_rows[i][4]:
                if self.raw_rows[i][1] == -1:
                    if i == self.cursel:
                        color_string = '{!black,red,bold!}'
                    else:
                        color_string = '{!red,black!}'
                else:
                    if i == self.cursel:
                        color_string = '{!black,cyan,bold!}'
                    else:
                        color_string = '{!cyan,black!}'

            elif i == self.cursel:
                if self.raw_rows[i][0] in self.marked:
                    color_string = '{!blue,white,bold!}'
                else:
                    color_string = '{!black,white,bold!}'
            elif self.raw_rows[i][0] in self.marked:
                color_string = '{!white,blue,bold!}'

            self.add_string(off, color_string + row)
            off += 1

            if off > self.rows - 2:
                break

        if not component.get('ConsoleUI').is_active_mode(self):
            return

        self.stdscr.noutrefresh()

        if self.popup:
            self.popup.refresh()

        curses.doupdate()

    def back_to_overview(self):
        self.parent_mode.go_top = False
        component.get('ConsoleUI').set_mode(self.parent_mode.mode_name)

    def _perform_action(self):
        if self.cursel < len(self.listing_dirs):
            self._enter_dir()
        else:
            s = self.raw_rows[self.cursel][0]
            if s not in self.marked:
                self.last_mark = self.cursel
            self.marked.add(s)
            self._show_add_dialog()

    def _enter_dir(self):
        # Enter currently selected directory
        dirname = self.raw_rows[self.cursel][0]
        new_dir = self.path_stack_pos >= len(self.path_stack)
        new_dir = new_dir or (dirname != self.path_stack[self.path_stack_pos])
        if new_dir:
            self.path_stack = self.path_stack[: self.path_stack_pos]
            self.path_stack.append(dirname)

        path = os.path.join(*self.path_stack[: self.path_stack_pos + 1])

        if not os.access(path, os.R_OK):
            self.path_stack = self.path_stack[: self.path_stack_pos]
            self.popup = MessagePopup(
                self, 'Error', '{!error!}Access denied: %s' % path
            )
            self.__refresh_listing()
            return

        self.path_stack_pos += 1

        self.view_offset = 0
        self.cursel = 0
        self.last_mark = -1
        self.marked = set()

        self.__refresh_listing()

    def _show_add_dialog(self):
        def _do_add(result, **kwargs):
            ress = {'succ': 0, 'fail': 0, 'total': len(self.marked), 'fmsg': []}

            def fail_cb(msg, t_file, ress):
                log.debug('failed to add torrent: %s: %s', t_file, msg)
                ress['fail'] += 1
                ress['fmsg'].append(f'{{!input!}} * {t_file}: {{!error!}}{msg}')
                if (ress['succ'] + ress['fail']) >= ress['total']:
                    report_add_status(
                        component.get('TorrentList'),
                        ress['succ'],
                        ress['fail'],
                        ress['fmsg'],
                    )

            def success_cb(tid, t_file, ress):
                if tid:
                    log.debug('added torrent: %s (%s)', t_file, tid)
                    ress['succ'] += 1
                    if (ress['succ'] + ress['fail']) >= ress['total']:
                        report_add_status(
                            component.get('TorrentList'),
                            ress['succ'],
                            ress['fail'],
                            ress['fmsg'],
                        )
                else:
                    fail_cb('Already in session (probably)', t_file, ress)

            for m in self.marked:
                filename = m
                directory = os.path.join(*self.path_stack[: self.path_stack_pos])
                path = os.path.join(directory, filename)
                with open(path, 'rb') as _file:
                    filedump = b64encode(_file.read())
                t_options = {}
                if result['location']['value']:
                    t_options['download_location'] = result['location']['value']
                t_options['add_paused'] = result['add_paused']['value']

                d = client.core.add_torrent_file_async(filename, filedump, t_options)
                d.addCallback(success_cb, filename, ress)
                d.addErrback(fail_cb, filename, ress)

            self.console_config['addtorrents']['last_path'] = os.path.join(
                *self.path_stack[: self.path_stack_pos]
            )
            self.console_config.save()

            self.back_to_overview()

        config = component.get('ConsoleUI').coreconfig
        if config['add_paused']:
            ap = 0
        else:
            ap = 1
        self.popup = InputPopup(
            self, 'Add Torrents (Esc to cancel)', close_cb=_do_add, height_req=17
        )

        msg = 'Adding torrent files:'
        for i, m in enumerate(self.marked):
            name = m
            msg += '\n * {!input!}%s' % name
            if i == 5:
                if i < len(self.marked):
                    msg += '\n  {!red!}And %i more' % (len(self.marked) - 5)
                break
        self.popup.add_text(msg)
        self.popup.add_spaces(1)

        self.popup.add_text_input(
            'location', 'Download Folder:', config['download_location'], complete=True
        )
        self.popup.add_select_input(
            'add_paused', 'Add Paused:', ['Yes', 'No'], [True, False], ap
        )

    def _go_up(self):
        # Go up in directory hierarchy
        if self.path_stack_pos > 1:
            self.path_stack_pos -= 1

            self.view_offset = 0
            self.cursel = 0
            self.last_mark = -1
            self.marked = set()

            self.__refresh_listing()

    def read_input(self):
        c = self.stdscr.getch()

        if self.popup:
            if self.popup.handle_read(c):
                self.popup = None
            self.refresh()
            return

        if util.is_printable_chr(c):
            if chr(c) == 'Q':
                component.get('ConsoleUI').quit()
            elif chr(c) == 'q':
                self.back_to_overview()
                return

        # Navigate the torrent list
        if c == curses.KEY_UP:
            self.scroll_list_up(1)
        elif c == curses.KEY_PPAGE:
            self.scroll_list_up(self.rows // 2)
        elif c == curses.KEY_HOME:
            self.scroll_list_up(len(self.formatted_rows))
        elif c == curses.KEY_DOWN:
            self.scroll_list_down(1)
        elif c == curses.KEY_NPAGE:
            self.scroll_list_down(self.rows // 2)
        elif c == curses.KEY_END:
            self.scroll_list_down(len(self.formatted_rows))
        elif c == curses.KEY_RIGHT:
            if self.cursel < len(self.listing_dirs):
                self._enter_dir()
        elif c == curses.KEY_LEFT:
            self._go_up()
        elif c in [curses.KEY_ENTER, util.KEY_ENTER2]:
            self._perform_action()
        elif c == util.KEY_ESC:
            self.back_to_overview()
        else:
            if util.is_printable_chr(c):
                if chr(c) == 'h':
                    self.popup = MessagePopup(self, 'Help', HELP_STR, width_req=0.75)
                elif chr(c) == '>':
                    if self.sort_column == 'date':
                        self.reverse_sort = not self.reverse_sort
                    else:
                        self.sort_column = 'date'
                        self.reverse_sort = True
                    self.__sort_rows()
                elif chr(c) == '<':
                    if self.sort_column == 'name':
                        self.reverse_sort = not self.reverse_sort
                    else:
                        self.sort_column = 'name'
                        self.reverse_sort = False
                    self.__sort_rows()
                elif chr(c) == 'm':
                    s = self.raw_rows[self.cursel][0]
                    if s in self.marked:
                        self.marked.remove(s)
                    else:
                        self.marked.add(s)

                    self.last_mark = self.cursel
                elif chr(c) == 'j':
                    self.scroll_list_down(1)
                elif chr(c) == 'k':
                    self.scroll_list_up(1)
                elif chr(c) == 'M':
                    if self.last_mark != -1:
                        if self.last_mark > self.cursel:
                            m = list(range(self.cursel, self.last_mark))
                        else:
                            m = list(range(self.last_mark, self.cursel + 1))

                        for i in m:
                            s = self.raw_rows[i][0]
                            self.marked.add(s)
                elif chr(c) == 'c':
                    self.marked.clear()

        self.refresh()



#
# Copyright (C) 2008 Andrew Resch <andrewresch@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import logging

import deluge.component as component
from deluge.ui.client import client

log = logging.getLogger(__name__)


class CoreConfig(component.Component):
    def __init__(self):
        log.debug('CoreConfig init..')
        component.Component.__init__(self, 'CoreConfig')
        self.config = {}

        def on_configvaluechanged_event(key, value):
            self.config[key] = value

        client.register_event_handler(
            'ConfigValueChangedEvent', on_configvaluechanged_event
        )

    def start(self):
        def on_get_config(config):
            self.config = config
            return config

        return client.core.get_config().addCallback(on_get_config)

    def stop(self):
        self.config = {}

    def __contains__(self, key):
        return key in self.config

    def __getitem__(self, key):
        return self.config[key]

    def __setitem__(self, key, value):
        client.core.set_config({key: value})

    def __getattr__(self, attr):
        # We treat this directly interacting with the dictionary
        return getattr(self.config, attr)



#
# Copyright (C) 2008-2009 Ido Abramovich <ido.deluge@gmail.com>
# Copyright (C) 2009 Andrew Resch <andrewresch@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#
import fnmatch
import logging
import os
import sys

import deluge.common
from deluge.argparserbase import ArgParserBase, DelugeTextHelpFormatter
from deluge.ui.ui import UI

log = logging.getLogger(__name__)

#
# Note: Cannot import from console.main here because it imports the twisted reactor.
# Console is imported from console/__init__.py loaded by the script entry points
# defined in setup.py
#


def load_commands(command_dir):
    def get_command(name):
        command = getattr(
            __import__(
                'deluge.ui.console.cmdline.commands.%s' % name, {}, {}, ['Command']
            ),
            'Command',
        )()
        command._name = name
        return command

    try:
        dir_list = fnmatch.filter(os.listdir(command_dir), '*.py')
    except OSError:
        return {}

    commands = []
    for filename in dir_list:
        if filename.startswith('_'):
            continue
        cmd = get_command(os.path.splitext(filename)[0])
        for cmd_name in [cmd._name] + cmd.aliases:
            commands.append((cmd_name, cmd))
    return dict(commands)


class LogStream:
    out = sys.stdout

    def write(self, data):
        self.out.write(data)

    def flush(self):
        self.out.flush()


class Console(UI):
    cmd_description = """Console or command-line user interface"""

    def __init__(self, *args, **kwargs):
        super().__init__('console', *args, log_stream=LogStream(), **kwargs)

        group = self.parser.add_argument_group(
            _('Console Options'),
            _(
                'These daemon connect options will be '
                'used for commands, or if console ui autoconnect is enabled.'
            ),
        )
        group.add_argument(
            '-d',
            '--daemon',
            metavar='<ip_addr>',
            dest='daemon_addr',
            help=_('Deluge daemon IP address to connect to (default 127.0.0.1)'),
            default='127.0.0.1',
        )
        group.add_argument(
            '-p',
            '--port',
            metavar='<port>',
            dest='daemon_port',
            type=int,
            help=_('Deluge daemon port to connect to (default 58846)'),
            default='58846',
        )
        group.add_argument(
            '-U',
            '--username',
            metavar='<user>',
            dest='daemon_user',
            help=_('Deluge daemon username to use when connecting'),
        )
        group.add_argument(
            '-P',
            '--password',
            metavar='<pass>',
            dest='daemon_pass',
            help=_('Deluge daemon password to use when connecting'),
        )
        # To properly print help message for the console commands ( e.g. deluge-console info -h),
        # we add a subparser for each command which will trigger the help/usage when given
        from deluge.ui.console.parser import (  # import here because (see top)
            ConsoleCommandParser,
        )

        self.console_parser = ConsoleCommandParser(
            parents=[self.parser],
            add_help=False,
            prog=self.parser.prog,
            description='Starts the Deluge console interface',
            formatter_class=lambda prog: DelugeTextHelpFormatter(
                prog, max_help_position=33, width=90
            ),
        )
        self.parser.subparser = self.console_parser
        self.console_parser.base_parser = self.parser
        subparsers = self.console_parser.add_subparsers(
            title=_('Console Commands'),
            help=_('Description'),
            description=_('The following console commands are available:'),
            metavar=_('Command'),
            dest='command',
        )
        from deluge.ui.console import UI_PATH  # Must import here

        self.console_cmds = load_commands(os.path.join(UI_PATH, 'cmdline', 'commands'))
        for cmd in sorted(self.console_cmds):
            self.console_cmds[cmd].add_subparser(subparsers)

    def start(self):
        if self.ui_args is None:
            # Started directly by deluge-console script so must find the UI args manually
            options, remaining = ArgParserBase(common_help=False).parse_known_args()
            self.ui_args = remaining

        i = self.console_parser.find_subcommand(args=self.ui_args)
        self.console_parser.subcommand = False
        self.parser.subcommand = False if i == -1 else True

        super().start(self.console_parser)
        from deluge.ui.console.main import ConsoleUI  # import here because (see top)

        def run(options):
            try:
                c = ConsoleUI(self.options, self.console_cmds, self.parser.log_stream)
                return c.start_ui()
            except Exception as ex:
                log.exception(ex)
                raise

        return deluge.common.run_profiled(
            run,
            self.options,
            output_file=self.options.profile,
            do_profile=self.options.profile,
        )



# SPDX-License-Identifier: GPL-3.0-or-later WITH GPL-3.0-linking-exception
import sys
import zipfile
from importlib.metadata import EntryPoints
from pathlib import Path

import pytest

from deluge.pluginmanagerbase import (
    PluginManagerBase,
    _load_plugin_metadata,
    _parse_pkg_info,
    _PluginDist,
    plugin_name_to_id,
)


def test_parse_pkg_info_metadata_1_x_no_description():
    """Metadata 1.x has no body fallback; Description must be empty string."""
    pkg_info = 'Metadata-Version: 1.0\nName: Foo\nVersion: 0.1\n'
    info = _parse_pkg_info(pkg_info)
    assert info['Description'] == ''


def test_parse_pkg_info_metadata_2_1():
    pkg_info = """Metadata-Version: 2.1
Name: AutoAdd
Version: 1.8
Summary: Monitors folders for .torrent files.
Home-page: http://dev.deluge-torrent.org/wiki/Plugins/AutoAdd
Author: Chase Sterling, Pedro Algarvio
Author-email: chase.sterling@gmail.com, pedro@algarvio.me
License: GPLv3
Platform: UNKNOWN

Monitors folders for .torrent files.
    """
    plugin_info = _parse_pkg_info(pkg_info)
    for value in plugin_info.values():
        assert value != ''
    result = 'Monitors folders for .torrent files.'
    assert plugin_info['Description'] == result


def test_scan_finds_specific_builtin_plugins(builtin_egg_info_dir):
    """Well-known built-in plugins must all be present."""
    expected = [
        'AutoAdd',
        'Blocklist',
        'Execute',
        'Extractor',
        'Label',
        'Notifications',
        'Scheduler',
        'Stats',
        'Toggle',
        'WebUi',
    ]
    pm = PluginManagerBase(
        'core.conf', 'deluge.plugin.core', plugin_dirs=[builtin_egg_info_dir]
    )
    assert sorted(pm.available_plugins) == expected


def test_plugin_info_all_values_are_strings(builtin_egg_info_dir):
    """All values from get_plugin_info must be strings, with Name and Version always set."""
    pm = PluginManagerBase(
        'core.conf', 'deluge.plugin.core', plugin_dirs=[builtin_egg_info_dir]
    )
    for plugin in pm.available_plugins:
        info = pm.get_plugin_info(plugin)
        for key, value in info.items():
            assert isinstance(key, str), f'{plugin}: key {key!r} is not str'
            assert isinstance(value, str), f'{plugin}: value for {key!r} is not str'
        assert info['Name'], f'{plugin}: Name is empty'
        assert info['Version'], f'{plugin}: Version is empty'


def test_plugin_info_invalid_name():
    """get_plugin_info for unknown plugin returns sentinel values."""
    pm = PluginManagerBase('core.conf', 'deluge.plugin.core', plugin_dirs=[])
    info = pm.get_plugin_info('NonExistentPlugin')
    assert info['Name'] == 'not available'
    assert info['Version'] == 'not available'


def test_scan_egg_zip(fake_egg):
    """A minimal .egg zip dropped in the plugins dir must be discovered."""
    plugins_dir, _ = fake_egg
    pm = PluginManagerBase('core.conf', 'deluge.plugin.core', plugin_dirs=[plugins_dir])
    assert 'FakePlugin' in pm.available_plugins


def test_scan_egg_link(tmp_path):
    """A .egg-link file pointing at a source dir must be discovered."""
    source_dir = tmp_path / 'MyPlugin'
    egg_info_dir = source_dir / 'MyPlugin.egg-info'
    egg_info_dir.mkdir(parents=True)
    (egg_info_dir / 'PKG-INFO').write_text(
        'Metadata-Version: 1.0\nName: MyPlugin\nVersion: 0.1\n'
    )
    (egg_info_dir / 'entry_points.txt').write_text(
        '[deluge.plugin.core]\nMyPlugin = my_plugin:CorePlugin\n'
    )

    plugins_dir = tmp_path / 'plugins'
    plugins_dir.mkdir()
    (plugins_dir / 'MyPlugin.egg-link').write_text(str(source_dir) + '\n.\n')

    pm = PluginManagerBase('core.conf', 'deluge.plugin.core', plugin_dirs=[plugins_dir])
    assert 'MyPlugin' in pm.available_plugins
    assert pm.get_plugin_info('MyPlugin')['Version'] == '0.1'
    assert pm._plugin_dist['myplugin'].entry_points.select(group='deluge.plugin.core')


def test_scan_unpacked_egg_dir(tmp_path):
    """An unpacked .egg directory (Ubuntu-style) with EGG-INFO/ must be discovered."""
    plugins_dir = tmp_path / 'plugins'
    plugins_dir.mkdir()
    egg_dir = plugins_dir / 'FakePlugin-0.1-py3.egg'
    egg_info = egg_dir / 'EGG-INFO'
    egg_info.mkdir(parents=True)
    (egg_info / 'PKG-INFO').write_text(
        'Metadata-Version: 1.0\nName: FakePlugin\nVersion: 0.1\n'
    )
    (egg_info / 'entry_points.txt').write_text(
        '[deluge.plugin.core]\nFakePlugin = fake_plugin:CorePlugin\n'
    )

    pm = PluginManagerBase('core.conf', 'deluge.plugin.core', plugin_dirs=[plugins_dir])
    assert 'FakePlugin' in pm.available_plugins
    assert pm.get_plugin_info('FakePlugin')['Version'] == '0.1'


def test_scan_egg_zip_plugin_info(fake_egg):
    """get_plugin_info returns correct metadata read from EGG-INFO/PKG-INFO inside a zip egg."""
    plugins_dir, _ = fake_egg
    pm = PluginManagerBase('core.conf', 'deluge.plugin.core', plugin_dirs=[plugins_dir])
    info = pm.get_plugin_info('FakePlugin')
    assert info['Name'] == 'FakePlugin'
    assert info['Version'] == '0.1'


def test_no_duplicate_plugins(fake_egg):
    """The same plugin discovered in multiple scan dirs must appear only once."""
    plugins_dir, _ = fake_egg
    # Pass the same dir twice to simulate the plugin appearing in multiple scan paths
    pm = PluginManagerBase(
        'core.conf', 'deluge.plugin.core', plugin_dirs=[plugins_dir, plugins_dir]
    )
    names = [n for n in pm.available_plugins if n == 'FakePlugin']
    assert len(names) == 1


def test_higher_version_wins(tmp_path):
    """When two eggs for the same plugin are found, the higher version must be kept."""

    def make_egg(directory, version):
        pkg_info = f'Metadata-Version: 1.0\nName: FakePlugin\nVersion: {version}\n'
        egg_path = directory / f'FakePlugin-{version}-py3.egg'
        with zipfile.ZipFile(egg_path, 'w') as zf:
            zf.writestr('EGG-INFO/PKG-INFO', pkg_info)
            zf.writestr('EGG-INFO/entry_points.txt', '')
        return egg_path

    old_dir = tmp_path / 'old'
    new_dir = tmp_path / 'new'
    old_dir.mkdir()
    new_dir.mkdir()
    make_egg(old_dir, '0.1')
    make_egg(new_dir, '0.2')

    pm = PluginManagerBase(
        'core.conf', 'deluge.plugin.core', plugin_dirs=[old_dir, new_dir]
    )
    assert pm.get_plugin_info('FakePlugin')['Version'] == '0.2'


@pytest.mark.parametrize(
    'name, expected',
    [
        ('AutoAdd', 'autoadd'),
        ('Auto Add', 'auto-add'),
        ('my-plugin', 'my-plugin'),
        ('my_plugin', 'my-plugin'),
        ('My Plugin', 'my-plugin'),
        ('Caf\u00e9', 'cafe'),  # NFKD strips combining accent
    ],
)
def test_plugin_name_to_id(name, expected):
    """plugin_name_to_id must return lowercase kebab-case ASCII."""
    assert plugin_name_to_id(name) == expected


def test_plugin_info_has_location(fake_egg):
    """get_plugin_info must include a non-empty Location string."""
    plugins_dir, _ = fake_egg
    pm = PluginManagerBase('core.conf', 'deluge.plugin.core', plugin_dirs=[plugins_dir])
    info = pm.get_plugin_info('FakePlugin')
    assert 'Location' in info
    assert info['Location']  # non-empty
    assert Path(info['Location']).exists()


def test_load_plugin_metadata_ignores_unknown_files(tmp_path):
    """_load_plugin_metadata must return None for files it doesn't recognise."""
    junk = tmp_path / 'something.txt'
    junk.write_text('irrelevant')
    assert _load_plugin_metadata(junk) is None


def test_deactivate_clears_module_cache(tmp_path):
    """deactivate() must evict the plugin package and its submodules from sys.modules."""
    dist = _PluginDist(
        name='FakePlugin',
        version='0.1',
        location=tmp_path,
        entry_points=EntryPoints(),
        pkg_info='',
    )
    dist._active = True

    # Simulate modules loaded when the plugin was active
    sys.modules['fakeplugin'] = object()
    sys.modules['fakeplugin.core'] = object()

    dist.deactivate()

    assert 'fakeplugin' not in sys.modules
    assert 'fakeplugin.core' not in sys.modules


def test_deactivate_evicts_entry_point_module(tmp_path):
    """deactivate() must use the entry point module name, not the dist id.

    For deluge plugins the top-level package (e.g. deluge_fakeplugin) differs
    from the distribution id (e.g. fakeplugin). The eviction must target the
    actual imported package, derived from the entry point value.
    """
    ep_txt = '[deluge.plugin.core]\nFakePlugin = deluge_fakeplugin:CorePlugin\n'
    eps = EntryPoints._from_text_for(ep_txt, None)
    dist = _PluginDist(
        name='FakePlugin',
        version='0.1',
        location=tmp_path,
        entry_points=eps,
        pkg_info='',
    )
    dist._active = True

    sys.modules['deluge_fakeplugin'] = object()
    sys.modules['deluge_fakeplugin.core'] = object()

    dist.deactivate()

    assert 'deluge_fakeplugin' not in sys.modules
    assert 'deluge_fakeplugin.core' not in sys.modules



#
# Copyright (C) 2011 Nick Lanham <nick@afternight.org>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import copy
import logging

import deluge.common
from deluge.i18n import setup_translation
from deluge.ui.common import TORRENT_DATA_FIELD
from deluge.ui.console.utils import format_utils

setup_translation()

log = logging.getLogger(__name__)

torrent_data_fields = copy.deepcopy(TORRENT_DATA_FIELD)

formatters = {
    'queue': format_utils.format_queue,
    'name': lambda a, b: b,
    'state': None,
    'tracker': None,
    'download_location': None,
    'owner': None,
    'progress_state': format_utils.format_progress,
    'progress': format_utils.format_progress,
    'size': format_utils.format_size,
    'downloaded': format_utils.format_size,
    'uploaded': format_utils.format_size,
    'remaining': format_utils.format_size,
    'ratio': format_utils.format_float,
    'avail': format_utils.format_float,
    'seeds_peers_ratio': format_utils.format_float,
    'download_speed': format_utils.format_speed,
    'upload_speed': format_utils.format_speed,
    'max_download_speed': format_utils.format_speed,
    'max_upload_speed': format_utils.format_speed,
    'peers': format_utils.format_seeds_peers,
    'seeds': format_utils.format_seeds_peers,
    'time_added': deluge.common.fdate,
    'seeding_time': format_utils.format_time,
    'active_time': format_utils.format_time,
    'time_since_transfer': format_utils.format_date_dash,
    'finished_time': deluge.common.ftime,
    'last_seen_complete': format_utils.format_date_never,
    'completed_time': format_utils.format_date_dash,
    'eta': format_utils.format_time,
    'pieces': format_utils.format_pieces,
}

for data_field in torrent_data_fields:
    torrent_data_fields[data_field]['formatter'] = formatters.get(data_field, str)


def get_column_value(name, state):
    col = torrent_data_fields[name]

    if col['formatter']:
        args = [state[key] for key in col['status']]
        return col['formatter'](*args)
    else:
        return state[col['status'][0]]


def get_required_fields(cols):
    fields = []
    for col in cols:
        fields.extend(torrent_data_fields[col]['status'])
    return fields



#
# Copyright (C) 2008-2009 Ido Abramovich <ido.deluge@gmail.com>
# Copyright (C) 2009 Andrew Resch <andrewresch@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

from twisted.internet import defer

import deluge.component as component
import deluge.log

from . import BaseCommand


class Command(BaseCommand):
    """Enable and disable debugging"""

    def add_arguments(self, parser):
        parser.add_argument(
            'state', metavar='<on|off>', choices=['on', 'off'], help=_('The new state')
        )

    def handle(self, options):
        if options.state == 'on':
            deluge.log.set_logger_level('debug')
        elif options.state == 'off':
            deluge.log.set_logger_level('error')
        else:
            component.get('ConsoleUI').write('{!error!}%s' % self.usage)

        return defer.succeed(True)

    def complete(self, text):
        return [x for x in ['on', 'off'] if x.startswith(text)]



class ACTION:
    PAUSE = 'pause'
    RESUME = 'resume'
    REANNOUNCE = 'update_tracker'
    EDIT_TRACKERS = 3
    RECHECK = 'force_recheck'
    REMOVE = 'remove_torrent'
    REMOVE_DATA = 6
    REMOVE_NODATA = 7
    DETAILS = 'torrent_details'
    MOVE_STORAGE = 'move_download_folder'
    QUEUE = 'queue'
    QUEUE_TOP = 'queue_top'
    QUEUE_UP = 'queue_up'
    QUEUE_DOWN = 'queue_down'
    QUEUE_BOTTOM = 'queue_bottom'
    TORRENT_OPTIONS = 'torrent_options'



#
# Copyright (C) 2011 Nick Lanham <nick@afternight.org>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import logging
import os.path

import deluge.component as component
from deluge.ui.client import client

from . import BaseCommand

log = logging.getLogger(__name__)


class Command(BaseCommand):
    """Move torrents' storage location"""

    def add_arguments(self, parser):
        parser.add_argument(
            'torrent_ids',
            metavar='<torrent-id>',
            nargs='+',
            help=_('One or more torrent ids'),
        )
        parser.add_argument(
            'path', metavar='<path>', help=_('The path to move the torrents to')
        )

    def handle(self, options):
        self.console = component.get('ConsoleUI')

        if os.path.exists(options.path) and not os.path.isdir(options.path):
            self.console.write(
                '{!error!}Cannot Move Download Folder: %s exists and is not a directory'
                % options.path
            )
            return

        ids = self.console.match_torrents(options.torrent_ids)
        names = [self.console.get_torrent_name(id_) for id_ in ids]

        def on_move(res):
            msg = 'Moved "{}" to {}'.format(', '.join(names), options.path)
            self.console.write(msg)
            log.info(msg)

        d = client.core.move_storage(ids, options.path)
        d.addCallback(on_move)
        return d

    def complete(self, line):
        line = os.path.abspath(os.path.expanduser(line))
        ret = []
        if os.path.exists(line):
            # This is a correct path, check to see if it's a directory
            if os.path.isdir(line):
                # Directory, so we need to show contents of directory
                # ret.extend(os.listdir(line))
                for f in os.listdir(line):
                    # Skip hidden
                    if f.startswith('.'):
                        continue
                    f = os.path.join(line, f)
                    if os.path.isdir(f):
                        f += '/'
                    ret.append(f)
            else:
                # This is a file, but we could be looking for another file that
                # shares a common prefix.
                for f in os.listdir(os.path.dirname(line)):
                    if f.startswith(os.path.split(line)[1]):
                        ret.append(os.path.join(os.path.dirname(line), f))
        else:
            # This path does not exist, so lets do a listdir on it's parent
            # and find any matches.
            ret = []
            if os.path.isdir(os.path.dirname(line)):
                for f in os.listdir(os.path.dirname(line)):
                    if f.startswith(os.path.split(line)[1]):
                        p = os.path.join(os.path.dirname(line), f)

                        if os.path.isdir(p):
                            p += '/'
                        ret.append(p)
        return ret



#
# Copyright (C) 2011 Nick Lanham <nick@afternight.org>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import logging

import deluge.component as component
from deluge.decorators import overrides
from deluge.ui.console.modes.basemode import BaseMode
from deluge.ui.console.utils.curses_util import is_printable_chr
from deluge.ui.console.widgets.popup import InputPopup, PopupsHandler, SelectablePopup
from deluge.ui.hostlist import HostList

try:
    import curses
except ImportError:
    pass

log = logging.getLogger(__name__)


class ConnectionManager(BaseMode, PopupsHandler):
    def __init__(self, stdscr, encoding=None):
        PopupsHandler.__init__(self)
        self.statuses = {}
        self.all_torrents = None
        self.hostlist = HostList()
        BaseMode.__init__(self, stdscr, encoding=encoding)

    def update_select_host_popup(self):
        if self.popup and not isinstance(self.popup, SelectablePopup):
            # Ignore MessagePopup on popup stack upon connect fail
            return

        selected_index = self.popup.current_selection() if self.popup else None

        popup = SelectablePopup(
            self,
            _('Select Host'),
            self._host_selected,
            border_off_west=1,
            active_wrap=True,
        )
        popup.add_header(
            "{!white,black,bold!}'Q'=%s, 'a'=%s, 'D'=%s"
            % (_('Quit'), _('Add Host'), _('Delete Host')),
            space_below=True,
        )

        for host_entry in self.hostlist.get_hosts_info():
            host_id, hostname, port, user = host_entry
            host_status = self.statuses.get(host_id)

            state = host_status[1] if host_status else 'Offline'
            state_color = 'green' if state in ('Online', 'Connected') else 'red'
            host_str = f'{hostname}:{port} [{state}]'

            args = {'data': host_id, 'foreground': state_color}
            popup.add_line(
                host_id, host_str, selectable=True, use_underline=True, **args
            )

        if selected_index:
            popup.set_selection(selected_index)

        self.push_popup(popup, clear=True)
        self.inlist = True
        self.refresh()

    def update_hosts_status(self):
        def on_host_status(status_info):
            self.statuses[status_info[0]] = status_info
            self.update_select_host_popup()

        for host_entry in self.hostlist.get_hosts_info():
            self.hostlist.get_host_status(host_entry[0]).addCallback(on_host_status)

    def _on_connected(self, result):
        def on_console_start(result):
            component.get('ConsoleUI').set_mode('TorrentList')

        d = component.get('ConsoleUI').start_console()
        d.addCallback(on_console_start)

    def _on_connect_fail(self, result):
        self.report_message('Failed to connect!', result.getErrorMessage())
        self.refresh()
        if hasattr(result, 'getTraceback'):
            log.exception(result)

    def _host_selected(self, selected_host, *args, **kwargs):
        if selected_host in self.statuses:
            d = self.hostlist.connect_host(selected_host)
            d.addCallback(self._on_connected)
            d.addErrback(self._on_connect_fail)

    def _do_add(self, result, **kwargs):
        if not result or kwargs.get('close', False):
            self.pop_popup()
        else:
            self.add_host(
                result['hostname']['value'],
                result['port']['value'],
                result['username']['value'],
                result['password']['value'],
            )

    def add_popup(self):
        self.inlist = False
        popup = InputPopup(
            self,
            _('Add Host (Up & Down arrows to navigate, Esc to cancel)'),
            border_off_north=1,
            border_off_east=1,
            close_cb=self._do_add,
        )
        popup.add_text_input('hostname', _('Hostname:'))
        popup.add_text_input('port', _('Port:'))
        popup.add_text_input('username', _('Username:'))
        popup.add_text_input('password', _('Password:'))
        self.push_popup(popup, clear=True)
        self.refresh()

    def add_host(self, hostname, port, username, password):
        log.info('Adding host: %s', hostname)
        if port.isdecimal():
            port = int(port)
        try:
            self.hostlist.add_host(hostname, port, username, password)
        except ValueError as ex:
            self.report_message(_('Error adding host'), f'{hostname}: {ex}')
        else:
            self.pop_popup()

    def delete_host(self, host_id):
        log.info('Deleting host: %s', host_id)
        self.hostlist.remove_host(host_id)
        self.update_select_host_popup()

    @overrides(component.Component)
    def start(self):
        self.refresh()

    @overrides(component.Component)
    def update(self):
        self.update_hosts_status()

    @overrides(BaseMode)
    def pause(self):
        self.pop_popup()
        BaseMode.pause(self)

    @overrides(BaseMode)
    def resume(self):
        BaseMode.resume(self)
        self.refresh()

    @overrides(BaseMode)
    def refresh(self):
        if self.mode_paused():
            return

        self.stdscr.erase()
        self.draw_statusbars()
        self.stdscr.noutrefresh()

        if not self.popup:
            self.update_select_host_popup()

        if self.popup:
            self.popup.refresh()

        curses.doupdate()

    @overrides(BaseMode)
    def on_resize(self, rows, cols):
        BaseMode.on_resize(self, rows, cols)

        if self.popup:
            self.popup.handle_resize()

        self.stdscr.erase()
        self.refresh()

    @overrides(BaseMode)
    def read_input(self):
        c = self.stdscr.getch()

        if is_printable_chr(c):
            if chr(c) == 'Q':
                component.get('ConsoleUI').quit()
            elif self.inlist:
                if chr(c) == 'q':
                    return
                elif chr(c) == 'D':
                    host_index = self.popup.current_selection()
                    host_id = self.popup.inputs[host_index].name
                    self.delete_host(host_id)
                    return
                elif chr(c) == 'a':
                    self.add_popup()
                    return

        if self.popup:
            if self.popup.handle_read(c) and self.popup.closed():
                self.pop_popup()
            self.refresh()



#
# Copyright (C) 2008-2009 Ido Abramovich <ido.deluge@gmail.com>
# Copyright (C) 2009 Andrew Resch <andrewresch@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import deluge.component as component
from deluge.ui.client import client

from . import BaseCommand


class Command(BaseCommand):
    """Update tracker for torrent(s)"""

    usage = 'update_tracker [ * | <torrent-id> [<torrent-id> ...] ]'
    aliases = ['reannounce']

    def add_arguments(self, parser):
        parser.add_argument(
            'torrent_ids',
            metavar='<torrent-id>',
            nargs='+',
            help='One or more torrent ids. "*" updates all torrents',
        )

    def handle(self, options):
        self.console = component.get('ConsoleUI')
        args = options.torrent_ids
        if options.torrent_ids[0] == '*':
            args = ['']

        torrent_ids = []
        for arg in args:
            torrent_ids.extend(self.console.match_torrent(arg))

        client.core.force_reannounce(torrent_ids)

    def complete(self, line):
        # We use the ConsoleUI torrent tab complete method
        return component.get('ConsoleUI').tab_complete_torrent(line)
